"""various job types for Theory() instances"""
import numpy as np

from AaronTools import addlogger
from AaronTools.finders import FlaggedAtoms
from AaronTools.theory import (
    GAUSSIAN_CONSTRAINTS,
    GAUSSIAN_COORDINATES,
    GAUSSIAN_ROUTE,
    ORCA_BLOCKS,
    ORCA_ROUTE,
    PSI4_BEFORE_GEOM,
    PSI4_COORDINATES,
    PSI4_JOB,
    PSI4_OPTKING,
    PSI4_SETTINGS,
    PSI4_BEFORE_JOB,
    SQM_QMMM,
    QCHEM_REM,
    QCHEM_SETTINGS,
)
from AaronTools.utils.utils import range_list, combine_dicts


def job_from_string(name, **kwargs):
    """
    return a job type for the given string
    valid names are:
        "opt" or "conf" with ".ts", ".transition_state", ".change", and ".con" extensions
            * .ts and .transition_state indicate a transition state optimization
            * .con indicates a constrained optimization - "constraints" should
              be in kwargs and the value should be a dictionary conformable with 
              the keyword of OptimizationJob
        "freq" with ".num" extensions
            * .num indicates a numerical frequnecy, as does kwargs["numerical"] = True
            kwargs can also have a "temperature" key
        "sp" or "energy" or "single-point"
        "force" or "gradient" with a ".num" extension
            * .num indicates a numerical frequnecy, as does kwargs["numerical"] = True
    """
    ext = None
    if "." in name:
        ext = name.split(".")[-1].lower()
    
    if any(name.lower().startswith(x) for x in ["opt", "conf"]):
        geom = kwargs.get("geometry", None)
        constraints = kwargs.get("constraints", None)

        if ext and (ext.startswith("ts") or ext.startswith("transition")):
            return OptimizationJob(transition_state=True, geometry=geom)
        
        if ext and ext.startswith("con") and constraints:
            return OptimizationJob(constraints=constraints)
        
        if ext and ext.startswith("change"):
            return OptimizationJob(
                constraints={"atoms": FlaggedAtoms()}, geometry=geom
            )
        return OptimizationJob(geometry=geom)
    
    if name.lower().startswith("freq"):
        numerical = kwargs.get("numerical", False)
        numerical = numerical or (ext and ext.startswith("num"))
        temperature = kwargs.get("temperature", 298.15)
        
        return FrequencyJob(numerical=numerical, temperature=temperature)
    
    if any(name.lower().startswith(x) for x in ["sp", "energy", "single-point"]):
        return SinglePointJob()
    
    if any(name.lower().startswith(x) for x in ["force", "gradient"]):
        numerical = kwargs.get("numerical", False)
        numerical = numerical or (ext and ext.startswith("num"))
        
        return ForceJob(numerical=numerical)
    
    raise ValueError("cannot determine job type from string: %s" % name)


class JobType:
    """
    parent class of all job types
    initialization keywords should be the same as attribute names
    """

    def __init__(self):
        pass

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False

        return self.__dict__ == other.__dict__

    def get_gaussian(self):
        """overwrite to return dict with GAUSSIAN_* keys"""
        pass

    def get_orca(self):
        """overwrite to return dict with ORCA_* keys"""
        pass

    def get_psi4(self):
        """overwrite to return dict with PSI4_* keys"""
        pass
    
    def get_sqm(self):
        """overwrite to return a dict with SQM_* keys"""
        pass
    
    def get_qchem(self):
        """overwrite to return a dict with QCHEM_* keys"""
        pass

    @staticmethod
    def resolve_error(error, theory, exec_type, geometry=None):
        """returns a copy of theory or modifies theory to attempt
        to resolve an error
        theory will be modified if it is not possible for the current theory
        to work for any job
        if the error is specific to the molecule and settings, theory will
        be copied, modified, and returned
        raises NotImplementedError if this job type has no fix for
        the error code
        error: error code (e.g. SCF_CONV; see fileIO ERROR)
        theory: Theory() instance
        exec_type: software program (i.e. gaussian, orca, etc.)
        
        optional kwargs:
        geometry: Geometry(), structure might be adjusted slightly if
            there are close contacts
        """

        if error.upper() == "CLASH":
            # if there is a clash, rotate substituents to mitigate clashing
            if geometry:
                geom_copy = geometry.copy()
                bad_subs = geom_copy.remove_clash()
                if not bad_subs:
                    geometry.update_structure(geom_copy.coords)
                    return None

        if error.upper() == "SCF_CONV":
            if exec_type.lower() == "gaussian":
                # SCF convergence issue, try different SCF algorithm
                out_theory = theory.copy()
                out_theory.kwargs = combine_dicts(
                     {
                        GAUSSIAN_ROUTE: {"scf": ["xqc"]}
                    },
                    out_theory.kwargs
                )
                return out_theory

            if exec_type.lower() == "orca":
                # SCF convergence issue, orca recommends ! SlowConv
                # and increasing SCF iterations
                out_theory = theory.copy()
                out_theory.kwargs = combine_dicts(
                     {
                        ORCA_ROUTE: ["SlowConv"],
                        ORCA_BLOCKS: {"scf": ["MaxIter 500"]}
                    },
                    out_theory.kwargs
                )
                return out_theory
        
            if exec_type.lower() == "psi4":
                out_theory = theory.copy()
                # if theory.charge < 0:
                #     # if the charge is negative, try adding two electrons
                #     # to get a guess that might work
                #     # as well as HF with a small basis set
                #     out_theory.kwargs = combine_dicts(
                #         {
                #             PSI4_BEFORE_JOB: {
                #                 [
                #                     "mol = get_active_molecule()",
                #                     "mol.set_molecular_charge(%i)" % (theory.charge + 2),
                #                     "nrg = energy($METHOD)",
                #                     "mol.set_molecular_charge(%i)" % theory.charge,
                #                 ]
                #             }
                #         },
                #         out_theory.kwargs,
                #     )
                # ^ this doesn't seem to help in general
                # do 500 iterations and dampen
                out_theory.kwargs = combine_dicts(
                    {
                        PSI4_SETTINGS: {
                            "damping_percentage": "15",
                            "maxiter": "500",
                        },
                    },
                    out_theory.kwargs,
                )
        
        raise NotImplementedError(
            "cannot fix %s errors for %s; check your input" % (error, exec_type)
        )


@addlogger
class OptimizationJob(JobType):
    """optimization job"""

    LOG = None

    def __init__(
        self,
        transition_state=False,
        constraints=None,
        geometry=None,
    ):
        """use transition_state=True to do a TS optimization
        constraints - dict with keys:
                      **** available for ORCA, Gaussian, and Psi4 ****
                      'atoms': atom identifiers/finders - atoms to constrain
                      'bonds': list(atom idenifiers/finders) - distances to constrain
                               each atom identifier in the list should result in exactly 2 atoms
                      'angles': list(atom idenifiers/finders) - 1-3 angles to constrain
                                each atom identifier should result in exactly 3 atoms
                      'torsions': list(atom identifiers/finders) - constrained dihedral angles
                                  each atom identifier should result in exactly 4 atoms
                      **** available for Gaussian and Psi4 ****
                      'x': list(atom identifiers/finders) - constrain the x coordinate of
                           these atoms
                      similarly, 'y' and 'z' are also accepted
                      'xgroup': list(tuple(list(atom idenifiers), x_val, hold)) -
                            constrain the x coordinate of these atoms to be the same
                            x_val - set x-coordinate to this value
                            hold - hold this value constant during the optimization
                                   if 'hold' is omitted, the value will not be held
                                   constant during the optimization
                            e.g. 'xgroup':[("1-6", 0, False), ("13-24", 3.25, False)]
                            this will keep atoms 1-6 and 13-24 in parallel planes, while also
                            allowing those planes to move
                      'ygroup' and 'zgroup' are also available, with analagous options

                      *** NOTE ***
                      for Gaussian, 'bonds', 'angles', and 'torsions' constraints cannot be mixed
                      with 'x', 'y', 'z', 'xgroup', 'ygroup', or 'zgroup' constraints
        geometry    - Geoemtry, will be set when using an AaronTools FileWriter"""
        super().__init__()

        self.transition_state = transition_state
        self.constraints = constraints
        self.geometry = geometry

    def get_gaussian(self):
        """returns a dict with keys: GAUSSIAN_ROUTE, GAUSSIAN_CONSTRAINTS"""
        if self.transition_state:
            out = {GAUSSIAN_ROUTE: {"Opt": ["ts"]}}
        else:
            out = {GAUSSIAN_ROUTE: {"Opt": []}}

        coords = self.geometry.coords.tolist()
        vars = []
        consts = []
        use_zmat = False

        group_count = 1

        if self.constraints is not None and any(
            self.constraints[key] for key in self.constraints.keys()
        ):
            for key in self.constraints:
                if key not in [
                    "x",
                    "y",
                    "z",
                    "xgroup",
                    "ygroup",
                    "zgroup",
                    "atoms",
                    "bonds",
                    "angles",
                    "torsions",
                ]:
                    raise NotImplementedError(
                        "%s constraints cannot be generated for Gaussian" % key
                    )
            out[GAUSSIAN_CONSTRAINTS] = []

            if "x" in self.constraints and self.constraints["x"]:
                x_atoms = self.geometry.find(self.constraints["x"])
                for i, atom in enumerate(self.geometry.atoms):
                    if atom in x_atoms:
                        var_name = "x%i" % (i + 1)
                        consts.append((var_name, atom.coords[0]))
                        coords[i] = [var_name, coords[i][1], coords[i][2]]

                if not use_zmat:
                    use_zmat = True
                    out[GAUSSIAN_ROUTE]["Opt"].append("Z-Matrix")

            if "y" in self.constraints and self.constraints["y"]:
                y_atoms = self.geometry.find(self.constraints["y"])
                for i, atom in enumerate(self.geometry.atoms):
                    if atom in y_atoms:
                        var_name = "y%i" % (i + 1)
                        consts.append((var_name, atom.coords[1]))
                        coords[i] = [coords[i][0], var_name, coords[i][2]]

                if not use_zmat:
                    use_zmat = True
                    out[GAUSSIAN_ROUTE]["Opt"].append("Z-Matrix")

            if "z" in self.constraints and self.constraints["z"]:
                z_atoms = self.geometry.find(self.constraints["z"])
                for i, atom in enumerate(self.geometry.atoms):
                    if atom in z_atoms:
                        var_name = "z%i" % (i + 1)
                        consts.append((var_name, atom.coords[2]))
                        coords[i] = [coords[i][0], coords[i][1], var_name]

                if not use_zmat:
                    use_zmat = True
                    out[GAUSSIAN_ROUTE]["Opt"].append("Z-Matrix")

            if "xgroup" in self.constraints:
                for constraint in self.constraints["xgroup"]:
                    if len(constraint) == 3:
                        finders, val, hold = constraint
                    else:
                        finders, val = constraint
                        hold = False
                    x_atoms = self.geometry.find(finders)
                    var_name = "gx%i" % group_count
                    group_count += 1
                    if hold:
                        consts.append([var_name, val])
                    else:
                        vars.append([var_name, val])
                    for i, atom in enumerate(self.geometry.atoms):
                        if atom in x_atoms:
                            coords[i] = [var_name, coords[i][1], coords[i][2]]

                if not use_zmat:
                    use_zmat = True
                    out[GAUSSIAN_ROUTE]["Opt"].append("Z-Matrix")

            if "ygroup" in self.constraints:
                for constraint in self.constraints["ygroup"]:
                    if len(constraint) == 3:
                        finders, val, hold = constraint
                    else:
                        finders, val = constraint
                        hold = False
                    y_atoms = self.geometry.find(finders)
                    var_name = "gy%i" % group_count
                    group_count += 1
                    if hold:
                        consts.append([var_name, val])
                    else:
                        vars.append([var_name, val])
                    for i, atom in enumerate(self.geometry.atoms):
                        if atom in y_atoms:
                            coords[i] = [coords[i][0], var_name, coords[i][2]]

                if not use_zmat:
                    use_zmat = True
                    out[GAUSSIAN_ROUTE]["Opt"].append("Z-Matrix")

            if "zgroup" in self.constraints:
                for constraint in self.constraints["zgroup"]:
                    if len(constraint) == 3:
                        finders, val, hold = constraint
                    else:
                        finders, val = constraint
                        hold = False
                    z_atoms = self.geometry.find(finders)
                    var_name = "gz%i" % group_count
                    group_count += 1
                    if hold:
                        consts.append([var_name, val])
                    else:
                        vars.append([var_name, val])
                    for i, atom in enumerate(self.geometry.atoms):
                        if atom in z_atoms:
                            coords[i] = [coords[i][0], coords[i][1], var_name]

                if not use_zmat:
                    use_zmat = True
                    out[GAUSSIAN_ROUTE]["Opt"].append("Z-Matrix")

            if "atoms" in self.constraints and self.constraints["atoms"]:
                try:
                    atoms = self.geometry.find(self.constraints["atoms"])
                except LookupError as e:
                    self.LOG.warning(e)
                    atoms = []
                for atom in atoms:
                    ndx = self.geometry.atoms.index(atom) + 1
                    if not use_zmat:
                        out[GAUSSIAN_CONSTRAINTS].append("%2i F" % ndx)
                    else:
                        for j, coord in enumerate(coords[ndx - 1]):
                            if isinstance(coord, str):
                                var_name = coord
                                for k, var in enumerate(vars):
                                    if var[0] == coord and not var[
                                        0
                                    ].startswith("g"):
                                        vars.pop(k)
                                        break
                                else:
                                    var_name = "%s%i" % (
                                        ["x", "y", "z"][j],
                                        ndx,
                                    )
                                    coords[ndx - 1][j] = var_name
                            else:
                                var_name = "%s%i" % (["x", "y", "z"][j], ndx)
                                coords[ndx - 1][j] = var_name
                            if not any(
                                const[0] == var_name for const in consts
                            ):
                                consts.append([var_name, atom.coords[j]])

                if not use_zmat:
                    if "ModRedundant" not in out[GAUSSIAN_ROUTE]["Opt"]:
                        out[GAUSSIAN_ROUTE]["Opt"].append("ModRedundant")

            if "bonds" in self.constraints:
                for constraint in self.constraints["bonds"]:
                    atom1, atom2 = self.geometry.find(constraint)
                    ndx1 = self.geometry.atoms.index(atom1) + 1
                    ndx2 = self.geometry.atoms.index(atom2) + 1
                    if not use_zmat:
                        out[GAUSSIAN_CONSTRAINTS].append(
                            "B %2i %2i F" % (ndx1, ndx2)
                        )
                    else:
                        raise NotImplementedError(
                            "cannot apply bond constraints when using Cartesian Z-Matrix, which"
                            + " is necessitated by x, y, or z constraints"
                        )

                if "ModRedundant" not in out[GAUSSIAN_ROUTE]["Opt"]:
                    out[GAUSSIAN_ROUTE]["Opt"].append("ModRedundant")

            if "angles" in self.constraints:
                for constraint in self.constraints["angles"]:
                    atom1, atom2, atom3 = self.geometry.find(constraint)
                    ndx1 = self.geometry.atoms.index(atom1) + 1
                    ndx2 = self.geometry.atoms.index(atom2) + 1
                    ndx3 = self.geometry.atoms.index(atom3) + 1
                    if not use_zmat:
                        out[GAUSSIAN_CONSTRAINTS].append(
                            "A %2i %2i %2i F" % (ndx1, ndx2, ndx3)
                        )
                    else:
                        raise NotImplementedError(
                            "cannot apply angle constraints when using Cartesian Z-Matrix, which"
                            + " is necessitated by x, y, or z constraints"
                        )

                if "ModRedundant" not in out[GAUSSIAN_ROUTE]["Opt"]:
                    out[GAUSSIAN_ROUTE]["Opt"].append("ModRedundant")

            if "torsions" in self.constraints:
                for constraint in self.constraints["torsions"]:
                    atom1, atom2, atom3, atom4 = self.geometry.find(constraint)
                    ndx1 = self.geometry.atoms.index(atom1) + 1
                    ndx2 = self.geometry.atoms.index(atom2) + 1
                    ndx3 = self.geometry.atoms.index(atom3) + 1
                    ndx4 = self.geometry.atoms.index(atom4) + 1
                    if not use_zmat:
                        out[GAUSSIAN_CONSTRAINTS].append(
                            "D %2i %2i %2i %2i F" % (ndx1, ndx2, ndx3, ndx4)
                        )
                    else:
                        raise NotImplementedError(
                            "cannot apply torsion constraints when using Cartesian Z-Matrix,"
                            + "which is necessitated by x, y, or z constraints"
                        )

                if "ModRedundant" not in out[GAUSSIAN_ROUTE]["Opt"]:
                    out[GAUSSIAN_ROUTE]["Opt"].append("ModRedundant")

        if consts or vars:
            for i, coord in enumerate(coords):
                for j, ax in enumerate(["x", "y", "z"]):
                    if isinstance(coord[j], float):
                        var_name = "%s%i" % (ax, i + 1)
                        vars.append((var_name, coord[j]))
                        coord[j] = var_name

        if consts or vars:
            for coord in coords:
                coord.insert(0, 0)

        out[GAUSSIAN_COORDINATES] = {
            "coords": coords,
            "variables": vars,
            "constants": consts,
        }

        return out

    def get_orca(self):
        """returns a dict with keys: ORCA_ROUTE, ORCA_BLOCKS"""
        if self.transition_state:
            out = {ORCA_ROUTE: ["OptTS"]}
        else:
            out = {ORCA_ROUTE: ["Opt"]}

        if self.constraints is not None and any(
            self.constraints[key] for key in self.constraints.keys()
        ):
            for key in self.constraints:
                if key not in [
                    "atoms",
                    "bonds",
                    "angles",
                    "torsions",
                ]:
                    raise NotImplementedError(
                        "%s constraints cannot be generated for ORCA" % key
                    )
            geom_block = "Constraints\n"
            if "atoms" in self.constraints:
                for constraint in self.constraints["atoms"]:
                    atom1 = self.geometry.find(constraint)[0]
                    ndx1 = self.geometry.atoms.index(atom1)
                    out_str = "        {C %2i C}\n" % (ndx1)
                    geom_block += out_str

            if "bonds" in self.constraints:
                for constraint in self.constraints["bonds"]:
                    atom1, atom2 = self.geometry.find(constraint)
                    ndx1 = self.geometry.atoms.index(atom1)
                    ndx2 = self.geometry.atoms.index(atom2)
                    out_str = "        {B %2i %2i C}\n" % (ndx1, ndx2)
                    geom_block += out_str
    
            if "angles" in self.constraints:
                for constraint in self.constraints["angles"]:
                    atom1, atom2, atom3 = self.geometry.find(constraint)
                    ndx1 = self.geometry.atoms.index(atom1)
                    ndx2 = self.geometry.atoms.index(atom2)
                    ndx3 = self.geometry.atoms.index(atom3)
                    out_str = "        {A %2i %2i %2i C}\n" % (ndx1, ndx2, ndx3)
                    geom_block += out_str

            if "torsions" in self.constraints:
                for constraint in self.constraints["torsions"]:
                    atom1, atom2, atom3, atom4 = self.geometry.find(constraint)
                    ndx1 = self.geometry.atoms.index(atom1)
                    ndx2 = self.geometry.atoms.index(atom2)
                    ndx3 = self.geometry.atoms.index(atom3)
                    ndx4 = self.geometry.atoms.index(atom4)
                    out_str = "        {D %2i %2i %2i %2i C}\n" % (
                        ndx1,
                        ndx2,
                        ndx3,
                        ndx4,
                    )
                    geom_block += out_str

            geom_block += "    end"
            out[ORCA_BLOCKS] = {"geom": [geom_block]}

        return out

    def get_psi4(self):
        """returns a dict with keys: PSI4_JOB, PSI4_OPTKING, PSI4_BEFORE_GEOM"""
        if self.transition_state:
            out = {
                PSI4_JOB: {"optimize": []},
                PSI4_SETTINGS: {"opt_type": ["ts"]},
            }
        else:
            out = {PSI4_JOB: {"optimize": []}}

        coords = self.geometry.coords.tolist()
        vars = []
        group_count = 1

        freeze_str = ""
        freeze_str += 'freeze_list = """\n'
        add_freeze_list = False

        # constraints
        if self.constraints is not None and any(
            [self.constraints[key] for key in self.constraints.keys()]
        ):
            for key in self.constraints:
                if key not in [
                    "x",
                    "y",
                    "z",
                    "xgroup",
                    "ygroup",
                    "zgroup",
                    "atoms",
                    "bonds",
                    "angles",
                    "torsions",
                ]:
                    raise NotImplementedError(
                        "%s constraints cannot be generated for Psi4" % key
                    )
            out[PSI4_OPTKING] = {}
            if (
                "x" in self.constraints
                and self.constraints["x"]
                and self.geometry is not None
            ):
                add_freeze_list = True
                atoms = self.geometry.find(self.constraints["x"])
                for atom in atoms:
                    freeze_str += "    %2i x\n" % (
                        self.geometry.atoms.index(atom) + 1
                    )

            if (
                "y" in self.constraints
                and self.constraints["y"]
                and self.geometry is not None
            ):
                add_freeze_list = True
                atoms = self.geometry.find(self.constraints["y"])
                for atom in atoms:
                    freeze_str += "    %2i y\n" % (
                        self.geometry.atoms.index(atom) + 1
                    )

            if (
                "z" in self.constraints
                and self.constraints["z"]
                and self.geometry is not None
            ):
                add_freeze_list = True
                atoms = self.geometry.find(self.constraints["z"])
                for atom in atoms:
                    freeze_str += "    %2i z\n" % (
                        self.geometry.atoms.index(atom) + 1
                    )

            if (
                "atoms" in self.constraints
                and self.constraints["atoms"]
                and self.geometry is not None
            ):
                add_freeze_list = True
                atoms = self.geometry.find(self.constraints["atoms"])
                for atom in atoms:
                    freeze_str += "    %2i xyz\n" % (
                        self.geometry.atoms.index(atom) + 1
                    )

            if "xgroup" in self.constraints:
                for constraint in self.constraints["xgroup"]:
                    if len(constraint) == 3:
                        finders, val, hold = constraint
                    else:
                        finders, val = constraint
                        hold = False
                    x_atoms = self.geometry.find(finders)
                    var_name = "gx%i" % group_count
                    group_count += 1
                    if hold:
                        add_freeze_list = True
                        for i, atom in enumerate(x_atoms):
                            freeze_str += "    %2i x\n" % (
                                self.geometry.atoms.index(atom) + 1
                            )
                            coords[i][0] = val
                    else:
                        vars.append([var_name, val, True])
                        for i, atom in enumerate(self.geometry.atoms):
                            if atom in x_atoms:
                                coords[i] = [
                                    coords[i][0],
                                    coords[i][1],
                                    var_name,
                                ]

            if "ygroup" in self.constraints:
                for constraint in self.constraints["ygroup"]:
                    if len(constraint) == 3:
                        finders, val, hold = constraint
                    else:
                        finders, val = constraint
                        hold = False
                    y_atoms = self.geometry.find(finders)
                    var_name = "gy%i" % group_count
                    group_count += 1
                    if hold:
                        add_freeze_list = True
                        for i, atom in enumerate(y_atoms):
                            freeze_str += "    %2i y\n" % (
                                self.geometry.atoms.index(atom) + 1
                            )
                            coords[i][1] = val
                    else:
                        vars.append([var_name, val, True])
                        for i, atom in enumerate(self.geometry.atoms):
                            if atom in y_atoms:
                                coords[i] = [
                                    coords[i][0],
                                    coords[i][1],
                                    var_name,
                                ]

            if "zgroup" in self.constraints:
                for constraint in self.constraints["zgroup"]:
                    if len(constraint) == 3:
                        finders, val, hold = constraint
                    else:
                        finders, val = constraint
                        hold = False
                    z_atoms = self.geometry.find(finders)
                    var_name = "gz%i" % group_count
                    group_count += 1
                    if hold:
                        add_freeze_list = True
                        for i, atom in enumerate(z_atoms):
                            freeze_str += "    %2i z\n" % (
                                self.geometry.atoms.index(atom) + 1
                            )
                            coords[i][2] = val
                    else:
                        vars.append([var_name, val, True])
                        for i, atom in enumerate(self.geometry.atoms):
                            if atom in z_atoms:
                                coords[i] = [
                                    coords[i][0],
                                    coords[i][1],
                                    var_name,
                                ]

            if add_freeze_list:
                freeze_str += '"""\n'
                freeze_str += "    \n"
                out[PSI4_BEFORE_GEOM] = [freeze_str]
                out[PSI4_OPTKING]["frozen_cartesian"] = ["$freeze_list"]

            if "bonds" in self.constraints:
                if self.constraints["bonds"] and self.geometry is not None:
                    out_str = '("\n'
                    for bond in self.constraints["bonds"]:
                        atom1, atom2 = self.geometry.find(bond)
                        out_str += "        %2i %2i\n" % (
                            self.geometry.atoms.index(atom1) + 1,
                            self.geometry.atoms.index(atom2) + 1,
                        )

                    out_str += '    ")\n'

                    out[PSI4_OPTKING]["frozen_distance"] = [out_str]

            if "angles" in self.constraints:
                if self.constraints["angles"] and self.geometry is not None:
                    out_str = '("\n'
                    for angle in self.constraints["angles"]:
                        atom1, atom2, atom3 = self.geometry.find(angle)
                        out_str += "        %2i %2i %2i\n" % (
                            self.geometry.atoms.index(atom1) + 1,
                            self.geometry.atoms.index(atom2) + 1,
                            self.geometry.atoms.index(atom3) + 1,
                        )

                    out_str += '    ")\n'

                    out[PSI4_OPTKING]["frozen_bend"] = [out_str]

            if "torsions" in self.constraints:
                if self.constraints["torsions"] and self.geometry is not None:
                    out_str += '("\n'
                    for torsion in self.constraints["torsions"]:
                        atom1, atom2, atom3, atom4 = self.geometry.find(
                            torsion
                        )
                        out_str += "        %2i %2i %2i %2i\n" % (
                            self.geometry.atoms.index(atom1) + 1,
                            self.geometry.atoms.index(atom2) + 1,
                            self.geometry.atoms.index(atom3) + 1,
                            self.geometry.atoms.index(atom4) + 1,
                        )

                    out_str += '    ")\n'

                    out[PSI4_OPTKING]["frozen_dihedral"] = [out_str]

        if vars:
            out[PSI4_COORDINATES] = {
                "coords": coords,
                "variables": vars,
            }

        return out

    def get_xcontrol(self, config, ref=None):
        """
        Generates xcontrol file constraints

        Returns: dict(xcontrol)
        """
        if ref is None:
            ref = "ref.xyz"
        xcontrol = ""
        # only put constraints in xcontrol file so this works with Crest also
        frozen = [i + 1 for i, a in enumerate(self.geometry) if bool(a.flag)]
        if frozen:
            frozen = range_list(frozen)
            xcontrol += "$fix\n"
            xcontrol += "  atoms: {}\n".format(frozen)
            xcontrol += "  freeze: {}\n".format(frozen)
        elif self.constraints:
            xcontrol += "$constrain\n"
            xcontrol += "  force constant={}\n".format(
                config["Job"].get("constrain_force", fallback="0.5")
            )
            xcontrol += "  reference={}\n".format(
                config["Job"].get("constrain_ref", fallback=ref)
            )
            constrained = set([])
            for bond in self.constraints.get("bonds", []):
                bond = self.geometry.find(bond)
                constrained.update(bond)
                xcontrol += "  distance: {},{},auto\n".format(
                    *(self.geometry.atoms.index(c) + 1 for c in bond)
                )
            for angle in self.constraints.get("angles", []):
                angle = self.geometry.find(angle)
                constrained.update(angle)
                xcontrol += "  angle: {},{},{},auto\n".format(
                    *(self.geometry.atoms.index(c) + 1 for c in angle)
                )
            for dihedral in self.constraints.get("torsions", []):
                dihedral = self.geometry.find(dihedral)
                constrained.update(dihedral)
                xcontrol += "  dihedral: {},{},{},{},auto\n".format(
                    *(self.geometry.atoms.index(c) + 1 for c in dihedral)
                )
            relaxed = {
                i + 1
                for i, a in enumerate(self.geometry.atoms)
                if a not in constrained
            }
            relaxed = range_list(relaxed)
            xcontrol += "$metadyn\n"
            xcontrol += "  atoms: {}\n".format(relaxed)
        xcontrol += "$end\n"
        return xcontrol

    def get_sqm(self):
        """returns a dict(), warnings for optimization jobs"""
        warnings = []
        if self.transition_state:
            warnings.append("cannot do TS optimization with sqm")
        
        if self.constraints:
            warnings.append("cannot constrain sqm optimization")
        
        return dict(), warnings

    def get_qchem(self):
        if self.transition_state:
            out = {QCHEM_REM: {"JOB_TYPE": "TS"}}
        else:
            out = {QCHEM_REM: {"JOB_TYPE": "OPT"}}
        
        # constraints
        if self.constraints is not None and any(
            [self.constraints[key] for key in self.constraints.keys()]
        ):
            out[QCHEM_SETTINGS] = {"opt": []}
            constraints = None
            fixed = None

            x_atoms = []
            y_atoms = []
            z_atoms = []
            xyz_atoms = []

            if "x" in self.constraints:
                x_atoms = self.geometry.find(self.constraints["x"])


            if "y" in self.constraints:
                y_atoms = self.geometry.find(self.constraints["y"])


            if "z" in self.constraints:
                z_atoms = self.geometry.find(self.constraints["z"])

            if "atoms" in self.constraints:
                xyz_atoms = self.geometry.find(self.constraints["atoms"])

            if any([x_atoms, y_atoms, z_atoms, xyz_atoms]):
                fixed = "FIXED"
                for atom in x_atoms:
                    fixed += "\n        %i X" % (self.geometry.atoms.index(atom) + 1)
                    if atom in xyz_atoms:
                        fixed += "YZ"
                        continue
                    if atom in y_atoms:
                        fixed += "Y"
                    if atom in z_atoms:
                        fixed += "Z"
                
                for atom in y_atoms:
                    if atom in x_atoms:
                        continue
                    fixed += "\n        %i " % (self.geometry.atoms.index(atom) + 1)
                    if atom in xyz_atoms:
                        fixed += "XYZ"
                        continue
                    fixed += "Y"
                    if atom in z_atoms:
                        fixed += "Z"
                
                for atom in y_atoms:
                    if atom in x_atoms or atom in y_atoms:
                        continue
                    fixed += "\n        %i " % (self.geometry.atoms.index(atom) + 1)
                    if atom in xyz_atoms:
                        fixed += "XYZ"
                        continue
                    fixed += "Z"

                for atom in xyz_atoms:
                    if any(atom in l for l in [x_atoms, y_atoms, z_atoms]):
                        continue
                    fixed += "\n        %i XYZ" % (self.geometry.atoms.index(atom) + 1)

            if "bonds" in self.constraints:
                if constraints is None:
                    constraints = "CONSTRAINT\n"
                for bond in self.constraints["bonds"]:
                    atom1, atom2 = self.geometry.find(bond)
                    constraints += "        STRE %2i %2i %9.5f\n" % (
                        self.geometry.atoms.index(atom1) + 1,
                        self.geometry.atoms.index(atom2) + 1,
                        atom1.dist(atom2),
                    )
            
            if "angles" in self.constraints:
                if constraints is None:
                    constraints = "CONSTRAINT\n"
                for angle in self.constraints["angles"]:
                    atom1, atom2, atom3 = self.geometry.find(angle)
                    constraints += "        BEND %2i %2i %2i %9.5f\n" % (
                        self.geometry.atoms.index(atom1) + 1,
                        self.geometry.atoms.index(atom2) + 1,
                        self.geometry.atoms.index(atom3) + 1,
                        np.rad2deg(atom2.angle(atom1, atom3)),
                    )
            
            if "torsions" in self.constraints:
                if constraints is None:
                    constraints = "CONSTRAINT\n"
                for angle in self.constraints["torsions"]:
                    atom1, atom2, atom3, atom4 = self.geometry.find(angle)
                    constraints += "        TORS %2i %2i %2i %2i %9.5f\n" % (
                        self.geometry.atoms.index(atom1) + 1,
                        self.geometry.atoms.index(atom2) + 1,
                        self.geometry.atoms.index(atom3) + 1,
                        self.geometry.atoms.index(atom4) + 1,
                        np.rad2deg(
                            self.geometry.dihedral(
                                atom1, atom2, atom3, atom4,
                            )
                        ),
                    )
            
            
            if fixed:
                fixed += "\n    ENDFIXED"
                out[QCHEM_SETTINGS]["opt"].append(fixed)

            if constraints:
                constraints += "    ENDCONSTRAINT"
                out[QCHEM_SETTINGS]["opt"].append(constraints)
        
        return out

    @staticmethod
    def resolve_error(error, theory, exec_type, geometry=None):
        """
        resolves optimization-specific errors
        errors resolved by JobType take priority

        optional kwargs:
        geometry: Geometry(), structure might be adjusted slightly if
            the software had an issue with generating internal coordinates
        """
        try:
            return super(OptimizationJob, OptimizationJob).resolve_error(
                error, theory, exec_type, geometry=geometry
            )
        except NotImplementedError:
            pass
        
        if exec_type.lower() == "gaussian":
            if error.upper() == "CONV_LINK":
                # optimization out of steps, add more steps
                out_theory = theory.copy()
                out_theory.kwargs = combine_dicts(
                     {GAUSSIAN_ROUTE: {"opt": ["MaxCycles=300"]}}, out_theory.kwargs,
                )
                return out_theory
            
            if error.upper() == "FBX":
                # FormBX error, just restart the job
                # adjusting the geometry slightly can help
                if geometry:
                    coords = geometry.coords
                    scale = 1e-3
                    coords += scale * np.random.random_sample - scale / 2
                    geometry.update_structure(coords)
                return None
            
            if error.upper() == "REDUND":
                # internal coordinate error, just restart the job
                if geometry:
                    coords = geometry.coords
                    scale = 1e-3
                    coords += scale * np.random.random_sample - scale / 2
                    geometry.update_structure(coords)
                return None                
        
        if exec_type.lower() == "orca":
            if error.upper() == "OPT_CONV":
                # optimization out of steps, add more steps
                out_theory = theory.copy()
                out_theory.kwargs = combine_dicts(
                     {ORCA_BLOCKS: {"geom": ["MaxIter 300"]}}, out_theory.kwargs,
                )
                return out_theory
        
        if exec_type.lower() == "psi4":
            if error.upper() == "ICOORD":
                out_theory = theory.copy()
                out_theory.kwargs = combine_dicts(
                     {PSI4_OPTKING: {"opt_coordinates": "cartesian"}}, out_theory.kwargs,
                )
                return out_theory
        
        raise NotImplementedError(
            "cannot fix %s errors for %s; check your input" % (error, exec_type)
        )


class FrequencyJob(JobType):
    """frequnecy job"""

    def __init__(self, numerical=False, temperature=None):
        """
        temperature in K for thermochem info, defaults to 298.15 K
        """
        super().__init__()
        if temperature is None:
            temperature = 298.15
        self.numerical = numerical
        self.temperature = temperature

    def get_gaussian(self):
        """returns a dict with keys: GAUSSIAN_ROUTE"""
        out = {
            GAUSSIAN_ROUTE: {
                "Freq": ["temperature=%.2f" % float(self.temperature)]
            }
        }
        if self.numerical:
            out[GAUSSIAN_ROUTE]["Freq"].append("Numerical")

        return out

    def get_orca(self):
        """returns a dict with keys: ORCA_ROUTE"""
        out = {ORCA_BLOCKS: {"freq": ["Temp    %.2f" % self.temperature]}}
        if self.numerical:
            out[ORCA_ROUTE] = ["NumFreq"]
        else:
            out[ORCA_ROUTE] = ["Freq"]

        return out

    def get_psi4(self):
        """returns a dict with keys: PSI4_JOB"""
        out = {
            PSI4_JOB: {"frequencies": []},
            PSI4_SETTINGS: {"T": ["%.2f" % self.temperature]},
        }
        if self.numerical:
            out[PSI4_JOB]["frequencies"].append('dertype="gradient"')

        return out

    def get_sqm(self):
        raise NotImplementedError("cannot build frequnecy job input for sqm")

    def get_qchem(self):
        out = {QCHEM_REM: {"JOB_TYPE": "Freq"}}
        if self.numerical:
            out[QCHEM_REM]["FD_DERIVATIVE_TYPE"] = "1"

        return out

    @staticmethod
    def resolve_error(error, theory, exec_type, geometry=None):
        """
        resolves frequnecy-specific errors
        errors resolved by JobType take priority
        """
        try:
            return super(FrequencyJob, FrequencyJob).resolve_error(
                error, theory, exec_type, geometry=geometry
            )
        except NotImplementedError:
            pass
        
        if exec_type.lower() == "orca":
            if error.upper() == "NUMFREQ":
                # analytical derivatives are not available
                for job in theory.job_type:
                    if isinstance(job, FrequencyJob):
                        job.numerical = True
                return None
        
        raise NotImplementedError(
            "cannot fix %s errors for %s; check your input" % (error, exec_type)
        )


class SinglePointJob(JobType):
    """single point energy"""

    def get_gaussian(self):
        """returns a dict with keys: GAUSSIAN_ROUTE"""
        return {GAUSSIAN_ROUTE: {"SP": []}}

    def get_orca(self):
        """returns a dict with keys: ORCA_ROUTE"""
        return {ORCA_ROUTE: ["SP"]}

    def get_psi4(self):
        """returns a dict with keys: PSI4_JOB"""
        return {PSI4_JOB: {"energy": []}}
    
    def get_sqm(self):
        """returns a dict with keys: SQM_QMMM"""
        return {SQM_QMMM: {"maxcyc": ["0"]}}

    def get_qchem(self):
        out = {QCHEM_REM: {"JOB_TYPE": "SP"}}
        
        return out


class ForceJob(JobType):
    """force/gradient job"""

    def __init__(self, numerical=False):
        super().__init__()
        self.numerical = numerical

    def get_gaussian(self):
        """returns a dict with keys: GAUSSIAN_ROUTE"""
        out = {GAUSSIAN_ROUTE: {"force": []}}
        if self.numerical:
            out[GAUSSIAN_ROUTE]["force"].append("EnGrad")
        return out

    def get_orca(self):
        """returns a dict with keys: ORCA_ROUTE"""
        return {ORCA_ROUTE: ["NumGrad" if self.numerical else "EnGrad"]}

    def get_psi4(self):
        """returns a dict with keys: PSI4_JOB"""
        out = {PSI4_JOB: {"gradient": []}}
        if self.numerical:
            out[PSI4_JOB]["gradient"].append("dertype='energy'")
        return out

    def get_qchem(self):
        out = {QCHEM_REM: {"JOB_TYPE": "Force"}}
        
        return out



"""used to specify implicit solvent info for Theory()"""
from AaronTools import addlogger
from AaronTools.theory import (
    GAUSSIAN_ROUTE, ORCA_BLOCKS, ORCA_ROUTE, PSI4_SETTINGS, PSI4_SOLVENT
)


class ImplicitSolvent:
    """implicit solvent info"""

    KNOWN_GAUSSIAN_SOLVENTS = [
        "Water",
        "Acetonitrile",
        "Methanol",
        "Ethanol",
        "IsoQuinoline",
        "Quinoline",
        "Chloroform",
        "DiethylEther",
        "DichloroMethane",
        "DiChloroEthane",
        "CarbonTetraChloride",
        "Benzene",
        "Toluene",
        "ChloroBenzene",
        "NitroMethane",
        "Heptane",
        "CycloHexane",
        "Aniline",
        "Acetone",
        "TetraHydroFuran",
        "DiMethylSulfoxide",
        "Argon",
        "Krypton",
        "Xenon",
        "n-Octanol",
        "1,1,1-TriChloroEthane",
        "1,1,2-TriChloroEthane",
        "1,2,4-TriMethylBenzene",
        "1,2-DiBromoEthane",
        "1,2-EthaneDiol",
        "1,4-Dioxane",
        "1-Bromo-2-MethylPropane",
        "1-BromoOctane",
        "1-BromoPentane",
        "1-BromoPropane",
        "1-Butanol",
        "1-ChloroHexane",
        "1-ChloroPentane",
        "1-ChloroPropane",
        "1-Decanol",
        "1-FluoroOctane",
        "1-Heptanol",
        "1-Hexanol",
        "1-Hexene",
        "1-Hexyne",
        "1-IodoButane",
        "1-IodoHexaDecane",
        "1-IodoPentane",
        "1-IodoPropane",
        "1-NitroPropane",
        "1-Nonanol",
        "1-Pentanol",
        "1-Pentene",
        "1-Propanol",
        "2,2,2-TriFluoroEthanol",
        "2,2,4-TriMethylPentane",
        "2,4-DiMethylPentane",
        "2,4-DiMethylPyridine",
        "2,6-DiMethylPyridine",
        "2-BromoPropane",
        "2-Butanol",
        "2-ChloroButane",
        "2-Heptanone",
        "2-Hexanone",
        "2-MethoxyEthanol",
        "2-Methyl-1-Propanol",
        "2-Methyl-2-Propanol",
        "2-MethylPentane",
        "2-MethylPyridine",
        "2-NitroPropane",
        "2-Octanone",
        "2-Pentanone",
        "2-Propanol",
        "2-Propen-1-ol",
        "3-MethylPyridine",
        "3-Pentanone",
        "4-Heptanone",
        "4-Methyl-2-Pentanone",
        "4-MethylPyridine",
        "5-Nonanone",
        "AceticAcid",
        "AcetoPhenone",
        "a-ChloroToluene",
        "Anisole",
        "Benzaldehyde",
        "BenzoNitrile",
        "BenzylAlcohol",
        "BromoBenzene",
        "BromoEthane",
        "Bromoform",
        "Butanal",
        "ButanoicAcid",
        "Butanone",
        "ButanoNitrile",
        "ButylAmine",
        "ButylEthanoate",
        "CarbonDiSulfide",
        "Cis-1,2-DiMethylCycloHexane",
        "Cis-Decalin",
        "CycloHexanone",
        "CycloPentane",
        "CycloPentanol",
        "CycloPentanone",
        "Decalin-mixture",
        "DiBromoMethane",
        "DiButylEther",
        "DiEthylAmine",
        "DiEthylSulfide",
        "DiIodoMethane",
        "DiIsoPropylEther",
        "DiMethylDiSulfide",
        "DiPhenylEther",
        "DiPropylAmine",
        "E-1,2-DiChloroEthene",
        "E-2-Pentene",
        "EthaneThiol",
        "EthylBenzene",
        "EthylEthanoate",
        "EthylMethanoate",
        "EthylPhenylEther",
        "FluoroBenzene",
        "Formamide",
        "FormicAcid",
        "HexanoicAcid",
        "IodoBenzene",
        "IodoEthane",
        "IodoMethane",
        "IsoPropylBenzene",
        "m-Cresol",
        "Mesitylene",
        "MethylBenzoate",
        "MethylButanoate",
        "MethylCycloHexane",
        "MethylEthanoate",
        "MethylMethanoate",
        "MethylPropanoate",
        "m-Xylene",
        "n-ButylBenzene",
        "n-Decane",
        "n-Dodecane",
        "n-Hexadecane",
        "n-Hexane",
        "NitroBenzene",
        "NitroEthane",
        "n-MethylAniline",
        "n-MethylFormamide-mixture",
        "n,n-DiMethylAcetamide",
        "n,n-DiMethylFormamide",
        "n-Nonane",
        "n-Octane",
        "n-Pentadecane",
        "n-Pentane",
        "n-Undecane",
        "o-ChloroToluene",
        "o-Cresol",
        "o-DiChloroBenzene",
        "o-NitroToluene",
        "o-Xylene",
        "Pentanal",
        "PentanoicAcid",
        "PentylAmine",
        "PentylEthanoate",
        "PerFluoroBenzene",
        "p-IsoPropylToluene",
        "Propanal",
        "PropanoicAcid",
        "PropanoNitrile",
        "PropylAmine",
        "PropylEthanoate",
        "p-Xylene",
        "Pyridine",
        "sec-ButylBenzene",
        "tert-ButylBenzene",
        "TetraChloroEthene",
        "TetraHydroThiophene-s,s-dioxide",
        "Tetralin",
        "Thiophene",
        "Thiophenol",
        "trans-Decalin",
        "TriButylPhosphate",
        "TriChloroEthene",
        "TriEthylAmine",
        "Xylene-mixture",
        "Z-1,2-DiChloroEthene",
    ]

    KNOWN_GAUSSIAN_MODELS = [
        "SMD",
        "CPCM",
        "PCM",
        "DIPOLE",
        "IPCM",
        "ISODENSITY",
        "IEFPCM",
        "SCIPCM",
    ]

    KNOWN_ORCA_CPCM_SOLVENTS = [
        "Water",
        "Acetonitrile",
        "Acetone",
        "Ammonia",
        "Ethanol",
        "Methanol",
        "CH2Cl2",
        "CCl4",
        "DMF",
        "DMSO",
        "Pyridine",
        "THF",
        "Chloroform",
        "Hexane",
        "Benzene",
        "CycloHexane",
        "Octanol",
        "Toluene",
    ]

    KNOWN_ORCA_SMD_SOLVENTS = [
        "1,1,1-TRICHLOROETHANE",
        "CYCLOPENTANE",
        "1,1,2-TRICHLOROETHANE",
        "CYCLOPENTANOL",
        "1,2,4-TRIMETHYLBENZENE",
        "CYCLOPENTANONE",
        "1,2-DIBROMOETHANE",
        "DECALIN (CIS/TRANS MIXTURE)",
        "1,2-DICHLOROETHANE",
        "CIS-DECALIN",
        "1,2-ETHANEDIOL",
        "N-DECANE",
        "1,4-DIOXANE",
        "DIBROMOMETHANE",
        "1-BROMO-2-METHYLPROPANE",
        "DIBUTYLETHER",
        "1-BROMOOCTANE",
        "O-DICHLOROBENZENE",
        "1-BROMOPENTANE",
        "E-1,2-DICHLOROETHENE",
        "1-BROMOPROPANE",
        "Z-1,2-DICHLOROETHENE",
        "1-BUTANOL",
        "DICHLOROMETHANE",
        "1-CHLOROHEXANE",
        "DIETHYL ETHER",
        "1-CHLOROPENTANE",
        "DIETHYL SULFIDE",
        "1-CHLOROPROPANE",
        "DIETHYLAMINE",
        "1-DECANOL",
        "DIIODOMETHANE",
        "1-FLUOROOCTANE",
        "DIISOPROPYL ETHER",
        "1-HEPTANOL",
        "CIS-1,2-DIMETHYLCYCLOHEXANE",
        "1-HEXANOL",
        "DIMETHYL DISULFIDE",
        "1-HEXENE",
        "N,N-DIMETHYLACETAMIDE",
        "1-HEXYNE",
        "N,N-DIMETHYLFORMAMIDE",
        "DMF",
        "1-IODOBUTANE",
        "DIMETHYLSULFOXIDE",
        "DMSO",
        "1-IODOHEXADECANE",
        "DIPHENYLETHER",
        "1-IODOPENTANE",
        "DIPROPYLAMINE",
        "1-IODOPROPANE",
        "N-DODECANE",
        "1-NITROPROPANE",
        "ETHANETHIOL",
        "1-NONANOL",
        "ETHANOL",
        "1-OCTANOL",
        "ETHYL ETHANOATE",
        "1-PENTANOL",
        "ETHYL METHANOATE",
        "1-PENTENE",
        "ETHYL PHENYL ETHER",
        "1-PROPANOL",
        "ETHYLBENZENE",
        "2,2,2-TRIFLUOROETHANOL",
        "FLUOROBENZENE",
        "2,2,4-TRIMETHYLPENTANE",
        "FORMAMIDE",
        "2,4-DIMETHYLPENTANE",
        "FORMIC ACID",
        "2,4-DIMETHYLPYRIDINE",
        "N-HEPTANE",
        "2,6-DIMETHYLPYRIDINE",
        "N-HEXADECANE",
        "2-BROMOPROPANE",
        "N-HEXANE",
        "2-BUTANOL",
        "HEXANOIC ACID",
        "2-CHLOROBUTANE",
        "IODOBENZENE",
        "2-HEPTANONE",
        "IODOETHANE",
        "2-HEXANONE",
        "IODOMETHANE",
        "2-METHOXYETHANOL",
        "ISOPROPYLBENZENE",
        "2-METHYL-1-PROPANOL",
        "P-ISOPROPYLTOLUENE",
        "2-METHYL-2-PROPANOL",
        "MESITYLENE",
        "2-METHYLPENTANE",
        "METHANOL",
        "2-METHYLPYRIDINE",
        "METHYL BENZOATE",
        "2-NITROPROPANE",
        "METHYL BUTANOATE",
        "2-OCTANONE",
        "METHYL ETHANOATE",
        "2-PENTANONE",
        "METHYL METHANOATE",
        "2-PROPANOL",
        "METHYL PROPANOATE",
        "2-PROPEN-1-OL",
        "N-METHYLANILINE",
        "E-2-PENTENE",
        "METHYLCYCLOHEXANE",
        "3-METHYLPYRIDINE",
        "N-METHYLFORMAMIDE (E/Z MIXTURE)",
        "3-PENTANONE",
        "NITROBENZENE",
        "PhNO2",
        "4-HEPTANONE",
        "NITROETHANE",
        "4-METHYL-2-PENTANONE",
        "NITROMETHANE",
        "MeNO2",
        "4-METHYLPYRIDINE",
        "O-NITROTOLUENE",
        "5-NONANONE",
        "N-NONANE",
        "ACETIC ACID",
        "N-OCTANE",
        "ACETONE",
        "N-PENTADECANE",
        "ACETONITRILE",
        "MeCN",
        "PENTANAL",
        "ACETOPHENONE",
        "N-PENTANE",
        "ANILINE",
        "PENTANOIC ACID",
        "ANISOLE",
        "PENTYL ETHANOATE",
        "BENZALDEHYDE",
        "PENTYLAMINE",
        "BENZENE",
        "PERFLUOROBENZENE",
        "BENZONITRILE",
        "PROPANAL",
        "BENZYL ALCOHOL",
        "PROPANOIC ACID",
        "BROMOBENZENE",
        "PROPANONITRILE",
        "BROMOETHANE",
        "PROPYL ETHANOATE",
        "BROMOFORM",
        "PROPYLAMINE",
        "BUTANAL",
        "PYRIDINE",
        "BUTANOIC ACID",
        "TETRACHLOROETHENE",
        "BUTANONE",
        "TETRAHYDROFURAN",
        "THF",
        "BUTANONITRILE",
        "TETRAHYDROTHIOPHENE-S,S-DIOXIDE",
        "BUTYL ETHANOATE",
        "TETRALIN",
        "BUTYLAMINE",
        "THIOPHENE",
        "N-BUTYLBENZENE",
        "THIOPHENOL",
        "SEC-BUTYLBENZENE",
        "TOLUENE",
        "TERT-BUTYLBENZENE",
        "TRANS-DECALIN",
        "CARBON DISULFIDE",
        "TRIBUTYLPHOSPHATE",
        "CARBON TETRACHLORIDE",
        "CCl4",
        "TRICHLOROETHENE",
        "CHLOROBENZENE",
        "TRIETHYLAMINE",
        "CHLOROFORM",
        "N-UNDECANE",
        "A-CHLOROTOLUENE",
        "WATER",
        "O-CHLOROTOLUENE",
        "XYLENE (MIXTURE)",
        "M-CRESOL",
        "M-XYLENE",
        "O-CRESOL",
        "O-XYLENE",
        "CYCLOHEXANE",
        "P-XYLENE",
        "CYCLOHEXANONE",
    ]

    KNOWN_ORCA_MODELS = ["SMD", "CPCM", "C-PCM", "PCM"]

    KNOWN_PSI4_SOLVENTS = [
        "water",
        "propylene carbonate",
        "dimethylsolfoxide",
        "nitromethane",
        "aceotonitrile",
        "methanol",
        "ethanol",
        "acetone",
        "1,2-dichloroethane",
        "methylenechloride",
        "CH2Cl2",
        "tetrahydrofuran",
        "aniline",
        "chlorobenzene",
        "chloroform",
        "toluene",
        "1,4-dioxane",
        "benzene",
        "carbon tetrachloride",
        "cyclohexane",
        "n-heptane",
    ]
    
    KNOWN_PSI4_MODELS = ["CPCM", "IEFPCM"]

    LOG = None

    def __init__(self, solvent_model, solvent):
        self.solvent_model = solvent_model
        self.solvent = solvent

    def __repr__(self):
        return "%s(%s)" % (self.solvent_model.upper(), self.solvent.lower())

    def __eq__(self, other):
        return repr(self) == repr(other)

    def copy(self):
        new_dict = dict()
        for key, value in self.__dict__.items():
            try:
                new_dict[key] = value.copy()
            except AttributeError:
                new_dict[key] = value
                # ignore chimerax objects so seqcrow doesn't print a
                # warning when a geometry is copied
                if "chimerax" in value.__class__.__module__:
                    continue
                if value.__class__.__module__ != "builtins":
                    self.LOG.warning(
                        "No copy method for {}: in-place changes may occur".format(
                            type(value)
                        )
                    )
        
        return self.__class__(**new_dict)
 
    def get_gaussian(self):
        """returns dict() with solvent information for gaussian input files"""
        # need to check if solvent model is available
        warnings = []
        if self.solvent.lower() == "gas":
            # all gas, no solvent
            return (dict(), warnings)

        if not any(
                self.solvent_model.upper() == model for model in self.KNOWN_GAUSSIAN_MODELS
        ):
            warnings.append(
                "solvent model is not available in Gaussian: %s\nuse one of: %s"
                % (
                    self.solvent_model,
                    " ".join(
                        [
                            "SMD",
                            "CPCM",
                            "PCM",
                            "DIPOLE",
                            "IPCM",
                            "ISODENSITY",
                            "IEFPCM",
                            "SCIPCM",
                        ]
                    ),
                )
            )

        # check some orca solvent keywords and switch to gaussian ones
        solvent = self.solvent
        if solvent.lower() == "chcl2":
            solvent = "DiChloroMethane"
        elif solvent.lower() == "ccl4":
            solvent = "CarbonTetraChloride"
        elif solvent.lower() == "THF":
            solvent = "TetraHydroFuran"
        else:
            if not any(
                    solvent.lower() == gaussian_sol.lower()
                    for gaussian_sol in self.KNOWN_GAUSSIAN_SOLVENTS
            ):
                warnings.append(
                    "solvent is unknown to Gaussian: %s\n" % solvent +
                    "see AaronTools.theory.implicit_solvent.KNOWN_GAUSSIAN_SOLVENTS"
                )

        # route option: scrf(model,solvent=solvent solvent_model)
        return (
            {GAUSSIAN_ROUTE: {"scrf": [self.solvent_model, "solvent=%s" % solvent]}},
            warnings,
        )

    def get_orca(self):
        """returns dict() with solvent information for orca input files"""
        warnings = []
        if self.solvent.lower() == "gas":
            return (dict(), warnings)

        if not any(
                self.solvent_model.upper() == model for model in self.KNOWN_ORCA_MODELS
        ):
            warnings.append(
                "solvent model is not available in ORCA: %s\nuse CPCM or SMD"
                % self.solvent_model
            )

        out = {}
        cpcm = True
        # route option: CPCM(solvent name)
        # if using smd, add block %cpcm smd true end
        if self.solvent_model.upper() == "SMD":
            cpcm = False
            out[ORCA_BLOCKS] = {"cpcm": ["smd    true"]}

        solvent = self.solvent
        # orca has different solvents for cpcm and smd...
        # check both lists, might be able to switch a gaussian keyword to the correct orca one
        if cpcm:
            if solvent.lower() == "dichloromethane":
                solvent = "CH2Cl2"
            elif solvent.lower() == "carbontetrachloride":
                solvent = "CCl4"
            elif solvent.lower() == "tetrahydrofuran":
                solvent = "THF"
            else:
                if not any(
                        solvent.lower() == orca_sol.lower()
                        for orca_sol in self.KNOWN_ORCA_CPCM_SOLVENTS
                ):
                    warnings.append(
                        "solvent is unknown to ORCA: %s\n" % solvent +
                        "see AaronTools.theory.implicit_solvent.KNOWN_ORCA_CPCM_SOLVENTS"
                    )

        else:
            # TODO: look for gaussian/orca pcm solvent names that need to change
            if not any(
                    solvent.lower() == orca_sol.lower()
                    for orca_sol in self.KNOWN_ORCA_SMD_SOLVENTS
            ):
                warnings.append(
                    "solvent is unknown to ORCA: %s\n" % solvent +
                    "see AaronTools.theory.implicit_solvent.KNOWN_ORCA_SMD_SOLVENTS"
                )

        out[ORCA_ROUTE] = ["CPCM(%s)" % solvent]

        return (out, warnings)

    def get_psi4(self):
        """returns dict() with solvent information for psi4 input files"""

        warnings = []

        if self.solvent.lower() == "gas":
            return (dict(), warnings)

        if not any(
                self.solvent_model.upper() == model for model in self.KNOWN_ORCA_MODELS
        ):
            warnings.append(
                "solvent model is not available in ORCA: %s\nuse CPCM or SMD"
                % self.solvent_model
            )

        out = {}
        # route option: CPCM(solvent name)
        # if using smd, add block %cpcm smd true end
        out[PSI4_SETTINGS] = {"pcm": "true"}

        solvent = self.solvent
        # orca has different solvents for cpcm and smd...
        # check both lists, might be able to switch a gaussian keyword to the correct orca one
        if solvent.lower() == "dichloromethane":
            solvent = "CH2Cl2"
        elif solvent.lower() == "carbontetrachloride":
            solvent = "carbon tetrachloride"
        elif solvent.lower() == "thf":
            solvent = "tetrahydrofuran"
        else:
            if not any(
                    solvent.lower() == orca_sol.lower()
                    for orca_sol in self.KNOWN_ORCA_CPCM_SOLVENTS
            ):
                warnings.append(
                    "solvent may be unknown to Psi4: %s\n" % solvent +
                    "see AaronTools.theory.implicit_solvent.KNOWN_PSI4_SOLVENTS"
                )

        out[PSI4_SOLVENT] = {
            "Units": "Angstrom",
            "Medium":
                [
                    "SolverType = %s" % self.solvent_model,
                    "Solvent = \"%s\"" % solvent,
                ],
            "Mode": "Implicit"
        }

        return (out, warnings)


    def get_qchem(self):
        """returns dict() with solvent information for qchem input files"""

        warnings = []

        if self.solvent.lower() == "gas":
            return (dict(), warnings)

        raise NotImplementedError(
            "cannot create solvent info for Q-Chem input files"
        )



#!/usr/bin/env python3
import json
import os
import unittest

import numpy as np
from AaronTools import addlogger
from AaronTools.config import Config
from AaronTools.test import TestWithTimer, prefix
from AaronTools.theory import Theory, ImplicitSolvent, OptimizationJob


@addlogger
class TestConfig(TestWithTimer):
    LOG = None
    USER_SPECIFIC = ["metadata", "infile", (".*", ".*_dir")]
    config_list = [
        ("blank.ini", None),
        ("HOH.ini", None),
    ]
    theory_1 = os.path.join(prefix, "test_files", "theory_1.ini")
    theory_2 = os.path.join(prefix, "test_files", "theory_2.ini")

    def test_init(self):
        for i, (config_name, config) in enumerate(TestConfig.config_list):
            config = Config(
                os.path.join(prefix, "test_files", config_name),
                quiet=True,
                skip_user_default=True,
            )
            TestConfig.config_list[i] = config_name, config

            test = config.as_dict(skip=self.USER_SPECIFIC)
            ref_name = os.path.join(
                prefix, "ref_files", config_name.replace(".ini", "_init.json")
            )
            # with open(ref_name, "w") as f:
            #     json.dump(test, f, indent=2)

            # need this to make sure python->json stuff is consistent,
            # eg: json doesn't distinguish between tuples and lists
            test = json.loads(json.dumps(test))
            self.LOG.debug(json.dumps(test, indent=2))
            with open(ref_name, "r") as f:
                ref = json.load(f)
            self.maxDiff = None
            self.assertDictEqual(ref, test, msg=config_name)

    def test_parse_changes(self):
        ###
        # substitutions with single point of connections
        ###
        simple = Config(
            os.path.join(prefix, "test_files", "simple_subs.ini"),
            quiet=True,
            skip_user_default=True,
        )
        test = json.loads(json.dumps(simple._changes))
        ref_name = simple.infile.replace(".ini", "_changes.json")
        ref_name = ref_name.replace("test_files", "ref_files")
        # with open(ref_name, "w") as f:
        #     json.dump(simple._changes, f, indent=2)
        with open(ref_name, "r") as f:
            ref = json.load(f)
        self.assertDictEqual(ref, test, msg=simple.infile)

        ###
        # for ring-fusing substitutions (multiple connection points)
        ###
        rings = Config(
            os.path.join(prefix, "test_files", "ring_subs.ini"),
            quiet=True,
            skip_user_default=True,
        )
        test = json.loads(json.dumps(rings._changes))
        ref_name = rings.infile.replace(".ini", "_changes.json")
        ref_name = ref_name.replace("test_files", "ref_files")
        # with open(ref_name, "w") as f:
        #     json.dump(rings._changes, f, indent=2)
        with open(ref_name, "r") as f:
            ref = json.load(f)
        self.assertDictEqual(ref, test, msg=simple.infile)

    def test_for_loop(self):
        for_loop = Config(
            os.path.join(prefix, "test_files", "for_loop.ini"),
            quiet=True,
            skip_user_default=True,
        )
        structure_list = for_loop.get_template()
        for structure, kind in structure_list:
            dihedral = round(
                np.rad2deg(structure.dihedral("3", "1", "2", "6")), 6
            )
            if dihedral < 0:
                dihedral += 360
            self.assertEqual(float(structure.name.split(".")[-1]), dihedral)

    def test_call_on_suffix(self):
        config = Config(
            os.path.join(prefix, "test_files", "structure_suffix.ini"),
            quiet=True,
            skip_user_default=True,
        )

        structure_list = config.get_template()
        # for structure, kind in structure_list:
        #     print(structure, kind)

    def test_theory(self):
        config = Config(
            TestConfig.theory_1,
            quiet=True,
            skip_user_default=True,
        )
        geoms = config.get_template()
        geom = geoms[0][0]
        ref = Theory(
            method="b3lyp",
            basis="6-31G",
            solvent=ImplicitSolvent("PCM", "water"),
            empirical_dispersion="D3",
            job_type=OptimizationJob(geometry=geom),
            grid="(99, 590)",
            geometry=geom,
        )
        test = config.get_theory(geom)
        # print("\n", geom.write(outfile=False, theory=ref, style="gaussian"))
        
        self.assertEqual(ref, test)
        
        
        config = Config(
            TestConfig.theory_2,
            quiet=True,
            skip_user_default=True,
        )
        geoms = config.get_template()
        geom = geoms[0][0]
        ref = Theory(
            method="b3lyp",
            basis="zora-def2-tzvp aux J sarc Pt sarc-zora-tzvp",
            job_type="energy",
            geometry=geom,
            simple=["zora"],
        )
        # print("\n", geom.write(outfile=False, theory=ref, style="orca"))
        test = config.get_theory(geom)
        self.assertEqual(ref, test)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestConfig("test_init"))
    suite.addTest(TestConfig("test_parse_changes"))
    suite.addTest(TestConfig("test_for_loop"))
    suite.addTest(TestConfig("test_call_on_suffix"))
    suite.addTest(TestConfig("test_theory"))
    return suite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python3

import argparse
import os
import sys

import numpy as np
from AaronTools.geometry import Geometry
from AaronTools.fileIO import read_types, FileReader
from AaronTools.utils.utils import glob_files, get_filename


remove_frag_parser = argparse.ArgumentParser(
    description="remove a fragment from a molecule",
    formatter_class=argparse.RawTextHelpFormatter,
)

remove_frag_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file"
)

remove_frag_parser.add_argument(
    "-if", "--input-format",
    type=str,
    default=None,
    dest="input_format",
    choices=read_types,
    help="file format of input - xyz is assumed if input is stdin",
)

remove_frag_parser.add_argument(
    "-o",
    "--output",
    type=str,
    default=False,
    required=False,
    metavar="output destination",
    dest="outfile",
    help="output destination\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: stdout",
)

remove_frag_parser.add_argument(
    "-t",
    "--targets",
    type=str,
    required=True,
    dest="target",
    help="fragment atom connected to the rest of the molecule (1-indexed)",
)

remove_frag_parser.add_argument(
    "-k",
    "--keep-group",
    type=str,
    required=False,
    default=None,
    dest="avoid",
    help="atom on the molecule that is connected to the fragment being removed\nDefault: longest fragment",
)

remove_frag_parser.add_argument(
    "-a",
    "--add-hydrogen",
    action="store_true",
    required=False,
    default=False,
    dest="add_H",
    help="add hydrogen to cap where the fragment was removed",
)

args = remove_frag_parser.parse_args()

for infile in glob_files(args.infile, parser=remove_frag_parser):
    if isinstance(infile, str):
        if args.input_format is not None:
            f = FileReader((infile, args.input_format[0], infile))
        else:
            f = FileReader(infile)
    else:
        if args.input_format is not None:
            f = FileReader(("from stdin", args.input_format[0], infile))
        else:
            f = FileReader(("from stdin", "xyz", infile))

    geom = Geometry(f)
    
    for atom in geom.find(args.target):
        if atom in geom.atoms:
            geom.remove_fragment(atom, avoid=args.avoid, add_H=args.add_H)
            if not args.add_H:
                geom -= atom
    
    if args.outfile:
        outfile = args.outfile
        if "$INFILE" in outfile:
            outfile = outfile.replace("$INFILE", get_filename(infile))
        geom.write(append=False, outfile=outfile)
    else:
        s = geom.write(outfile=False)
        print(s)



#!/usr/bin/env python3

import sys
import numpy as np
import argparse

from AaronTools.geometry import Geometry
from AaronTools.fileIO import FileReader, read_types
from AaronTools.utils.utils import get_filename, glob_files

def four_atoms_and_a_float(vals):
    """check to see if argument is four numbers and a float"""
    err = argparse.ArgumentTypeError(
        "error with atom/dihedral specification: %s\n" % " ".join(vals) +
        "expected -s/-c int int int int float\n" +
        "example: -c 4 5 6 7 1.337"
    )
    out = []
    if len(vals) != 5:
        raise err
    for v in vals[:-1]:
        try:
            out.append(int(v))
        except ValueError:
            raise err

        if int(v) != float(v):
            raise err

    try:
        out.append(float(vals[-1]))
    except ValueError:
        raise err

    return out


dihedral_parser = argparse.ArgumentParser(
    description="measure or modify torsional angles",
    formatter_class=argparse.RawTextHelpFormatter
)

dihedral_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file"
)

dihedral_parser.add_argument(
    "-if", "--input-format",
    type=str,
    choices=read_types,
    default=None,
    dest="input_format",
    help="file format of input - xyz is assumed if input is stdin"
)

dihedral_parser.add_argument(
    "-m", "--measure",
    metavar=("atom1", "atom2", "atom3", "atom4"),
    action="append",
    type=str,
    nargs=4,
    default=[],
    required=False,
    dest="measure",
    help="measure and print the torsional angle (1-indexed)"
)

dihedral_parser.add_argument(
    "-c", "--change",
    metavar=("atom1", "atom2", "atom3", "atom4", "increment"),
    action="append",
    type=str,
    nargs=5,
    default=[],
    required=False,
    dest="change",
    help="change torsional angle by the amount specified"
)

dihedral_parser.add_argument(
    "-s", "--set",
    metavar=("atom1", "atom2", "atom3", "atom4", "angle"),
    action="append",
    type=str,
    nargs=5,
    default=[],
    required=False,
    dest="set_ang",
    help="set dihedral to the amount specified"
)

dihedral_parser.add_argument(
    "-r", "--radians",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="radians",
    help="work with radians instead of degrees"
)

dihedral_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    required=False,
    dest="outfile",
    metavar="output destination",
    help="output destination\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: stdout"
)

args = dihedral_parser.parse_args()

for f in glob_files(args.infile, dihedral_parser):
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format, None))
        else:
            infile = FileReader(f)
    else:
        if args.input_format is not None:
            infile = FileReader(("from stdin", args.input_format, f))
        else:
            infile = FileReader(("from stdin", "xyz", f))

    geom = Geometry(infile)

    # set dihedral to specified value
    for dihedral in args.set_ang:
        vals = four_atoms_and_a_float(dihedral)
        a1 = geom.find(str(vals[0]))[0]
        a2 = geom.find(str(vals[1]))[0]
        a3 = geom.find(str(vals[2]))[0]
        a4 = geom.find(str(vals[3]))[0]
        geom.change_dihedral(a1, a2, a3, a4, vals[4], radians=args.radians, adjust=False, as_group=True)

    #change dihedral by specified amount
    for dihedral in args.change:
        vals = four_atoms_and_a_float(dihedral)
        a1 = geom.find(str(vals[0]))[0]
        a2 = geom.find(str(vals[1]))[0]
        a3 = geom.find(str(vals[2]))[0]
        a4 = geom.find(str(vals[3]))[0]
        geom.change_dihedral(a1, a2, a3, a4, vals[4], radians=args.radians, adjust=True, as_group=True)

    #print specified dihedrals
    if len(args.infile) > 1:
        out = "%s\t" % f
    else:
        out = ""
    for dihedral in args.measure:
        a1 = geom.find(dihedral[0])[0]
        a2 = geom.find(dihedral[1])[0]
        a3 = geom.find(dihedral[2])[0]
        a4 = geom.find(dihedral[3])[0]
        val = geom.dihedral(a1, a2, a3, a4)
        if not args.radians:
            val *= 180 / np.pi

        out += "%f\n" % val

    out = out.rstrip()

    if len(args.set_ang) + len(args.change) > 0:
        if args.outfile:
            outfile = args.outfile
            if "$INFILE" in outfile:
                outfile = outfile.replace("$INFILE", get_filename(f))
            geom.write(append=True, outfile=outfile)
        else:
            print(geom.write(outfile=False))

    if len(args.measure) > 0:
        if not args.outfile:
            print(out)
        else:
            outfile = args.outfile
            if "$INFILE" in outfile:
                outfile = outfile.replace("$INFILE", get_filename(f))
            with open(outfile, "a") as f:
                f.write(out)


#!/usr/bin/env python
import json
import os
import unittest

import numpy as np
from AaronTools.comp_output import CompOutput
from AaronTools.component import Component
from AaronTools.fileIO import Frequency
from AaronTools.geometry import Geometry
from AaronTools.json_extension import ATDecoder, ATEncoder
from AaronTools.substituent import Substituent
from AaronTools.test import TestWithTimer, prefix
from AaronTools.theory import Theory, ImplicitSolvent, OptimizationJob


class TestJSON(TestWithTimer):
    small_mol = os.path.join(prefix, "test_files", "benzene_1-NO2_4-Cl.xyz")
    COCH3 = os.path.join(prefix, "test_files", "COCH3.xyz")
    lig = os.path.join(prefix, "test_files", "ligands", "R-Quinox-tBu3.xyz")
    cat = os.path.join(prefix, "test_files", "catalysts", "tm_multi-lig.xyz")
    log = os.path.join(prefix, "test_files", "normal.log")

    def json_tester(self, ref, equality_function, as_iter=False, **kwargs):
        # test to json
        test = json.dumps(ref, cls=ATEncoder, indent=4)
        # test from json
        test = json.loads(test, cls=ATDecoder)
        if as_iter:
            for r, t in zip(ref, test):
                equality_function(r, t, **kwargs)
        else:
            equality_function(ref, test, **kwargs)

    def atom_equal(self, ref, test, skip=[]):
        for key, val in ref.__dict__.items():
            if key in skip:
                continue
            try:
                if key == "coords":
                    self.assertTrue(
                        np.linalg.norm(ref.coords - test.coords) < 10 ** -15
                    )
                elif key == "tags":
                    self.assertSetEqual(ref.tags, test.tags)
                elif key in ["connected", "constraint"]:
                    self.assertEqual(
                        len(ref.__dict__[key]), len(test.__dict__[key])
                    )
                    for a, b in zip(
                        sorted(ref.__dict__[key]), sorted(test.__dict__[key])
                    ):
                        if key == "constraint":
                            self.assertEqual(a[1], b[1])
                            a = a[0]
                            b = b[0]
                        self.atom_equal(a, b, skip=["connected", "constraint"])
                else:
                    self.assertEqual(ref.__dict__[key], test.__dict__[key])
            except AssertionError as err:
                print(ref, ref.__dict__[key])
                print(test, test.__dict__[key])
                raise AssertionError("(key={}) {}".format(key, err))

    def geom_equal(self, ref, test, skip=[]):
        if "name" not in skip:
            self.assertEqual(ref.name, test.name)
        else:
            skip.remove("name")
        if "comment" not in skip:
            self.assertEqual(ref.comment, test.comment)
        for a, b in zip(
            sorted(ref.atoms, key=lambda x: float(x.name)),
            sorted(test.atoms, key=lambda x: float(x.name)),
        ):
            self.atom_equal(a, b, skip)

    def sub_equal(self, ref, test):
        self.assertEqual(ref.name, test.name)
        for a, b in zip(ref.atoms, test.atoms):
            self.atom_equal(a, b)
        self.assertEqual(ref.end, test.end)
        self.assertEqual(ref.conf_num, test.conf_num)
        self.assertEqual(ref.conf_angle, test.conf_angle)

    def component_equal(self, ref, test):
        self.geom_equal(ref, test, skip=["name"])
        # need to sort substituents to make sure the order is the same
        # use the name of the first atom in the substituent
        self.assertEqual(len(ref.substituents), len(test.substituents))
        for r, t in zip(
            sorted(ref.substituents, key=lambda x: int(x.atoms[0].name)),
            sorted(test.substituents, key=lambda x: int(x.atoms[0].name)),
        ):
            self.geom_equal(r, t, skip=["comment"])
        self.assertEqual(len(ref.backbone), len(test.backbone))
        for r, t in zip(sorted(ref.backbone), sorted(test.backbone)):
            self.atom_equal(r, t)
        self.assertEqual(len(ref.key_atoms), len(test.key_atoms))
        for r, t in zip(sorted(ref.key_atoms), sorted(test.key_atoms)):
            self.atom_equal(r, t)

    def catalyst_equal(self, ref, test):
        # only one of them needs the comment re-parsed, but
        # it shouldn't matter if we do both
        ref.parse_comment()
        test.parse_comment()
        self.geom_equal(ref, test)
        if ref.center is None:
            ref.detect_components()
        if test.center is None:
            test.detect_components()
        for r, t in zip(ref.center, test.center):
            self.atom_equal(r, t)
        for r, t in zip(sorted(ref.components), sorted(test.components)):
            self.component_equal(r, t)

    def comp_out_equal(self, ref, test):
        keys = [
            "geometry",
            "opts",
            "frequency",
            "archive",
            "gradient",
            "E_ZPVE",
            "ZPVE",
            "energy",
            "enthalpy",
            "free_energy",
            "grimme_g",
            "charge",
            "multiplicity",
            "mass",
            "temperature",
            "rotational_temperature",
            "rotational_symmetry_number",
            "error",
            "error_msg",
            "finished",
        ]
        for key in keys:
            rval = ref.__dict__[key]
            tval = test.__dict__[key]
            if key == "geometry":
                self.geom_equal(rval, tval)
            elif key == "opts":
                if rval is None:
                    continue
                for r, t in zip(rval, tval):
                    self.geom_equal(r, t)
            elif key == "frequency":
                self.freq_equal(rval, tval)
            elif key == "gradient":
                self.assertDictEqual(rval, tval)
            elif key == "rotational_temperature":
                self.assertListEqual(rval, tval)
            else:
                self.assertEqual(rval, tval)

    def freq_equal(self, ref, test):
        for r, t in zip(ref.data, test.data):
            self.assertEqual(r.frequency, t.frequency)
            self.assertEqual(r.intensity, t.intensity)
            self.assertTrue(np.linalg.norm(r.vector - t.vector) < 10 ** -12)
        self.assertListEqual(
            ref.imaginary_frequencies, test.imaginary_frequencies
        )
        self.assertListEqual(ref.real_frequencies, test.real_frequencies)
        self.assertEqual(ref.is_TS, test.is_TS)
        self.assertEqual(
            len(ref.by_frequency.keys()), len(test.by_frequency.keys())
        )
        for key in ref.by_frequency.keys():
            self.assertTrue(key in test.by_frequency)
            rv = ref.by_frequency[key]
            tv = test.by_frequency[key]
            self.assertEqual(rv["intensity"], tv["intensity"])
            self.assertTrue(
                np.linalg.norm(rv["vector"] - tv["vector"]) < 10 ** -12
            )

    def test_atom(self):
        mol = Geometry(TestJSON.small_mol)
        for i, a in enumerate(mol.atoms):
            a.add_tag("test" + str(i))
        self.json_tester(
            mol.atoms,
            self.atom_equal,
            as_iter=True,
            skip=["connected", "constraint"],
        )

    def test_geometry(self):
        mol = Geometry(TestJSON.small_mol)
        self.json_tester(mol, self.geom_equal)

    def test_substituent(self):
        sub = Substituent("COCH3")
        self.json_tester(sub, self.sub_equal)

    def test_component(self):
        lig = Component(TestJSON.lig)
        self.json_tester(lig, self.component_equal)

    def test_catalyst(self):
        cat = Geometry(TestJSON.cat)
        cat.fix_comment()
        cat.conf_num = 4
        self.json_tester(cat, self.catalyst_equal)

    def test_comp_output(self):
        log = CompOutput(TestJSON.log)
        self.json_tester(log, self.comp_out_equal)

    def test_frequency(self):
        log = CompOutput(TestJSON.log)
        freq = log.frequency
        self.json_tester(freq, self.freq_equal)

    def test_theory(self):
        # test with a bunch of different basis sets with different elements and
        # auxilliary basis sets
        ref_theory = Theory(
            method="DLPNO-CCSD(T)",
            basis="!H aug-cc-pvtz H cc-pvtz !H aux C aug-cc-pvtz H aux C cc-pvtz",
            simple=["TightSCF"],
        )
        ref_json = json.dumps(ref_theory, cls=ATEncoder)
        test_theory = json.loads(ref_json, cls=ATDecoder)
        self.assertEqual(ref_theory, test_theory)
        
        # test with many different theory objects and an ECP
        ref_theory = Theory(
            method="B3LYP",
            basis="!tm def2-SVP tm SDD",
            ecp="tm SDD",
            route={"opt": ["NoEigenTest"]},
            grid="SuperFineGrid",
            empirical_dispersion="D3",
            solvent=ImplicitSolvent("PCM", "benzene"),
            job_type=OptimizationJob(transition_state=True),
        )
        ref_json = json.dumps(ref_theory, cls=ATEncoder)
        test_theory = json.loads(ref_json, cls=ATDecoder)
        self.assertEqual(ref_theory, test_theory)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestJSON("test_atom"))
    suite.addTest(TestJSON("test_geometry"))
    suite.addTest(TestJSON("test_substituent"))
    suite.addTest(TestJSON("test_component"))
    suite.addTest(TestJSON("test_comp_output"))
    suite.addTest(TestJSON("test_catalyst"))
    suite.addTest(TestJSON("test_frequency"))
    suite.addTest(TestJSON("test_theory"))
    return suite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



"""For more complicated geometry manipulation and complex building"""
import os
import re

import numpy as np

from AaronTools.const import (
    AARONLIB,
    AARONTOOLS,
    BONDI_RADII,
    ELEMENTS,
    VDW_RADII,
)
from AaronTools.fileIO import read_types
from AaronTools.finders import BondedTo, CloserTo, NotAny
from AaronTools.geometry import Geometry
from AaronTools.substituent import Substituent
from AaronTools.utils.utils import perp_vector


class Component(Geometry):
    """
    Attributes:
    :name:          str
    :comment:       str
    :atoms:         list(Atom)
    :other:         dict()
    :substituents:  list(Substituent) substituents detected
    :backbone:      list(Atom) the backbone atoms
    :key_atoms:     list(Atom) the atoms used for mapping
    """

    AARON_LIBS = os.path.join(AARONLIB, "Ligands")
    BUILTIN = os.path.join(AARONTOOLS, "Ligands")
    FROM_SUBSTITUENTS = set([])

    def __init__(
        self,
        structure,
        name="",
        comment=None,
        tag=None,
        to_center=None,
        key_atoms=None,
        detect_backbone=True,
    ):
        """
        comp is either a file, a geometry, or an atom list
        """
        super().__init__()
        self.name = name
        self.comment = comment
        self.other = {}
        self.substituents = []
        self.backbone = None
        self.key_atoms = []

        if isinstance(structure, str) and not os.access(structure, os.R_OK):
            for ext in read_types:
                if structure.endswith(".%s" % ext):
                    structure = structure[: -(1 + len(ext))]

            for lib in [Component.AARON_LIBS, Component.BUILTIN]:
                if not os.path.exists(lib):
                    continue
                flig = None
                for f in os.listdir(lib):
                    name, ext = os.path.splitext(f)
                    if not any(".%s" % x == ext for x in read_types):
                        continue

                    match = structure == name
                    if match:
                        flig = os.path.join(lib, f)
                        break

                if flig:
                    break
            else:
                try:
                    structure = Substituent(structure)
                    Component.FROM_SUBSTITUENTS.add(structure.name)
                    self.__init__(structure, comment="K:1")
                    return
                except Exception:
                    raise FileNotFoundError(
                        "Cannot find ligand in library:", structure
                    )
            structure = flig
        super().__init__(structure, name, comment)

        if tag is not None:
            for a in self.atoms:
                a.add_tag(tag)

        self.other = self.parse_comment()
        try:
            self.key_atoms = self.find("key")
        except LookupError:
            if "key_atoms" in self.other:
                self.key_atoms = [
                    self.atoms[i] for i in self.other["key_atoms"]
                ]
        if key_atoms is not None:
            self.key_atoms = self.find(key_atoms)
        for a in self.key_atoms:
            a.tags.add("key")
        if detect_backbone:
            self.detect_backbone(to_center)
        self.rebuild()

    def __lt__(self, other):
        if len(self) != len(other):
            return len(self) < len(other)
        for a, b in zip(sorted(self.atoms), sorted(other.atoms)):
            if a < b:
                return True
        return False

    @classmethod
    def list(
        cls,
        name_regex=None,
        coordinating_elements=None,
        denticity=None,
        include_ext=False,
    ):
        names = []
        for lib in [cls.AARON_LIBS, cls.BUILTIN]:
            if not os.path.exists(lib):
                continue
            for f in os.listdir(lib):
                name, ext = os.path.splitext(f)
                if not any(".%s" % x == ext for x in read_types):
                    continue

                if name in names:
                    continue

                name_ok = True
                elements_ok = True
                denticity_ok = True

                if (
                    name_regex is not None
                    and re.search(name_regex, name, re.IGNORECASE) is None
                ):
                    name_ok = False

                if coordinating_elements is not None:
                    geom = Geometry(
                        os.path.join(lib, name + ext),
                        refresh_connected=False,
                        refresh_ranks=False,
                    )
                    # geom = cls(name)
                    elements = [
                        geom.atoms[i].element for i in geom.other["key_atoms"]
                    ]
                    if not all(
                        elements.count(x) == coordinating_elements.count(x)
                        for x in coordinating_elements
                    ) or not all(
                        coordinating_elements.count(x) == elements.count(x)
                        for x in elements
                    ):
                        elements_ok = False

                if denticity is not None:
                    geom = cls(name)
                    if len(geom.find("key")) != denticity:
                        denticity_ok = False

                if name_ok and elements_ok and denticity_ok:
                    if include_ext:
                        names.append(name + ext)
                    else:
                        names.append(name)

        return names + sorted(cls.FROM_SUBSTITUENTS)

    def c2_symmetric(self, to_center=None, tolerance=0.1):
        """determine if center-key atom axis is a C2 axis"""
        # determine ranks
        ranks = self.canonical_rank(
            update=False,
            break_ties=False,
            invariant=False,
        )
        # remove the rank of atoms that are along the c2 axis
        ranks_off_c2_axis = []
        if to_center is None:
            center = np.zeros(3)
        else:
            center = self.COM(to_center)

        v = self.COM(self.key_atoms) - center
        v /= np.linalg.norm(v)

        for atom, rank in zip(self.atoms, ranks):
            dist_along_v = np.dot(atom.coords - center, v)
            if (
                abs(np.linalg.norm(atom.coords - center) - dist_along_v)
                < tolerance
            ):
                continue

            ranks_off_c2_axis.append(rank)

        return all([ranks.count(x) % 2 == 0 for x in set(ranks_off_c2_axis)])

    def sterimol(self, to_center=None, bisect_L=False, **kwargs):
        """
        calculate ligand sterimol parameters for the ligand
        to_center - atom the ligand is coordinated to
        bisect_L - L axis will bisect (or analogous for higher denticity
                   ligands) the L-M-L angle
                   Default - center to centroid of key atoms
        **kwargs - arguments passed to Geometry.sterimol
        """
        if to_center is not None:
            center = self.find(to_center)
        else:
            center = self.find(
                [BondedTo(atom) for atom in self.key_atoms], NotAny(self.atoms)
            )

        if len(center) != 1:
            raise TypeError(
                "wrong number of center atoms specified;\n"
                "expected 1, got %i" % len(center)
            )
        center = center[0]

        if bisect_L:
            L_axis = np.zeros(3)
            for atom in self.key_atoms:
                v = center.bond(atom)
                v /= np.linalg.norm(v)
                v /= len(self.key_atoms)
                L_axis += v
        else:
            L_axis = self.COM(self.key_atoms) - center.coords
            L_axis /= np.linalg.norm(L_axis)

        return super().sterimol(L_axis, center, self.atoms, **kwargs)

    def copy(self, atoms=None, name=None, comment=None):
        rv = super().copy()
        return Component(rv)

    def rebuild(self):
        sub_atoms = []
        for sub in sorted(self.substituents):
            tmp = [sub.atoms[0]]
            tmp += sorted(sub.atoms[1:])
            for t in tmp:
                if t in sub_atoms:
                    continue
                if self.backbone and t in self.backbone:
                    continue
                sub_atoms += [t]
        if self.backbone is None:
            self.backbone = [a for a in self.atoms if a not in sub_atoms]
        self.backbone = sorted(self.backbone)
        self.atoms = self.backbone + sub_atoms

    def get_frag_list(self, targets=None, max_order=None):
        """
        find fragments connected by only one bond
        (both fragments contain no overlapping atoms)
        """
        if targets:
            atoms = self.find(targets)
        else:
            atoms = self.atoms
        frag_list = []
        for i, a in enumerate(atoms[:-1]):
            for b in atoms[i + 1 :]:
                if b not in a.connected:
                    continue

                frag_a = self.get_fragment(a, b)
                frag_b = self.get_fragment(b, a)
                if len(frag_a) == len(frag_b) and sorted(
                    frag_a, key=lambda x: ELEMENTS.index(x.element)
                ) == sorted(frag_b, key=lambda x: ELEMENTS.index(x.element)):
                    continue

                if len(frag_a) == 1 and frag_a[0].element == "H":
                    continue
                if len(frag_b) == 1 and frag_b[0].element == "H":
                    continue

                if max_order is not None and a.bond_order(b) > max_order:
                    continue

                if (frag_a, a, b) not in frag_list:
                    frag_list += [(frag_a, a, b)]
                if (frag_b, b, a) not in frag_list:
                    frag_list += [(frag_b, b, a)]
        return frag_list

    def detect_backbone(self, to_center=None):
        """
        Detects backbone and substituents attached to backbone
        Will tag atoms as 'backbone' or by substituent name

        :to_center:   the atoms connected to the metal/active center
        """
        # we must remove any tags already made
        for a in self.atoms:
            a.tags.discard("backbone")
        self.backbone = []
        if self.substituents is not None:
            for sub in self.substituents:
                for a in sub.atoms:
                    a.tags.discard(sub.name)
        self.substituents = []

        # get all possible fragments connected by one bond
        frag_list = self.get_frag_list()

        # get atoms connected to center
        if to_center is not None:
            to_center = self.find(to_center)
        else:
            try:
                to_center = self.find("key")
            except LookupError:
                to_center = []
            try:
                center = self.find("center")
                to_center += list(c.connected for c in center)
            except LookupError:
                center = []

        new_tags = {}  # hold atom tag options until assignment determined
        subs_found = {}  # for testing which sub assignment is best
        sub_atoms = set([])  # holds atoms assigned to substituents
        for frag_tup in sorted(frag_list, key=lambda x: len(x[0])):
            frag, start, end = frag_tup
            if frag[0] != start:
                frag = self.reorder(start=start, targets=frag)[0]
            # if frag contains atoms from to_center, it's part of backbone
            is_backbone = False
            for a in frag:
                if to_center and a in to_center:
                    is_backbone = True
                    break
            # skip substituent assignment if part of backbone
            if is_backbone:
                continue
            # try to find fragment in substituent library
            try:
                sub = Substituent(frag, end=end)
            except LookupError:
                continue
            if not to_center and len(frag) > len(self.atoms) - len(sub_atoms):
                break
            # save atoms and tags if found
            sub_atoms = sub_atoms.union(set(frag))
            subs_found[sub.name] = len(sub.atoms)
            for a in sub.atoms:
                if a in new_tags:
                    new_tags[a] += [sub.name]
                else:
                    new_tags[a] = [sub.name]
            # save substituent
            self.substituents += [sub]

        # tag substituents
        for a in new_tags:
            tags = new_tags[a]
            if len(tags) > 1:
                # if multiple substituent assignments possible,
                # want to keep the largest one (eg: tBu instead of Me)
                sub_length = []
                for t in tags:
                    sub_length += [subs_found[t]]
                max_length = max(sub_length)
                if max_length < 0:
                    max_length = min(sub_length)
                keep = sub_length.index(max_length)
                a.add_tag(tags[keep])
            else:
                a.add_tag(tags[0])

        # tag backbone
        for a in set(self.atoms) - set(sub_atoms):
            a.add_tag("backbone")
            self.backbone += [a]
        if not self.backbone:
            self.backbone = None
        return

    def capped_backbone(self, to_center=None, as_copy=True):
        if as_copy:
            comp = self.copy()
        else:
            comp = self

        if comp.backbone is None:
            comp.detect_backbone()

        subs = []
        for sub in comp.substituents:
            subs += [comp.remove_fragment(sub.atoms, sub.end, ret_frag=True)]

        if as_copy:
            comp.substituents = None
            return comp, subs
        else:
            return subs

    def minimize_sub_torsion(self, geom=None, **kwargs):
        """
        rotates substituents to minimize LJ potential
        geom: calculate LJ potential between self and another geometry-like
              object, instead of just within self
        """
        if geom is None:
            geom = self

        if self.substituents is None:
            self.detect_backbone()

        return super().minimize_sub_torsion(geom, **kwargs)

    def sub_rotate(self, start, angle=None):
        start = self.find_exact(start)[0]
        for sub in self.substituents:
            if sub.atoms[0] == start:
                break
        end = sub.end
        if angle is None:
            angle = sub.conf_angle
        if not angle:
            return
        self.change_dihedral(
            start, end, angle, fix=4, adjust=True, as_group=True
        )

    def cone_angle(
        self, center=None, method="exact", return_cones=False, radii="umn"
    ):
        """
        returns cone angle in degrees
        center - Atom() that this component is coordinating
                 used as the apex of the cone
        method (str) can be:
            'Tolman' - Tolman cone angle for unsymmetric ligands
                       See J. Am. Chem. Soc. 1974, 96, 1, 53–60 (DOI: 10.1021/ja00808a009)
                       NOTE: this does not make assumptions about the geometry
                       NOTE: only works with monodentate and bidentate ligands
            'exact' - cone angle from Allen et. al.
                      See Bilbrey, J.A., Kazez, A.H., Locklin, J. and Allen, W.D.
                      (2013), Exact ligand cone angles. J. Comput. Chem., 34:
                      1189-1197. (DOI: 10.1002/jcc.23217)
        return_cones - return cone apex, center of base, and base radius list
                       the sides of the cones will be 5 angstroms long
                       for Tolman cone angles, multiple cones will be returned, one for
                       each substituent coming off the coordinating atom
        radii: 'bondi' - Bondi vdW radii
               'umn'   - vdW radii from Mantina, Chamberlin, Valero, Cramer, and Truhlar
               dict() with elements as keys and radii as values
        """
        if method.lower() == "tolman":
            CITATION = "doi:10.1021/ja00808a009"
        elif method.lower() == "exact":
            CITATION = "doi:10.1002/jcc.23217"
        self.LOG.citation(CITATION)

        key = self.find("key")

        center = self.find_exact(center)[0]

        L_axis = self.COM(key) - center.coords
        L_axis /= np.linalg.norm(L_axis)

        if isinstance(radii, dict):
            radii_dict = radii
        elif radii.lower() == "bondi":
            radii_dict = BONDI_RADII
        elif radii.lower() == "umn":
            radii_dict = VDW_RADII

        # list of cone data for printing bild file or w/e
        cones = []

        if method.lower() == "tolman":
            total_angle = 0
            if len(key) > 2:
                raise NotImplementedError(
                    "Tolman cone angle not implemented for tridentate or more ligands"
                )

            elif len(key) == 2:
                key1, key2 = key
                try:
                    bridge_path = self.shortest_path(key1, key2)
                except LookupError:
                    bridge_path = False

            for key_atom in key:
                L_axis = key_atom.coords - center.coords
                L_axis /= np.linalg.norm(L_axis)
                bonded_atoms = self.find(BondedTo(key_atom))
                for bonded_atom in bonded_atoms:
                    frag = self.get_fragment(bonded_atom, key_atom)

                    use_bridge = False
                    if any(k in frag for k in key):
                        # fragment on bidentate ligands that connects to
                        # the other coordinating atom
                        k = self.find(frag, key)[0]
                        # the bridge might be part of a ring (e.g. BPY)
                        # to avoid double counting the bridge, check if the
                        # first atom in the fragment is the first atom on the
                        # path from one key atom to the other
                        if frag[0] in bridge_path:
                            use_bridge = True

                    if use_bridge:
                        # angle between one L-M bond and L-M-L bisecting vector
                        tolman_angle = center.angle(k, key_atom) / 2

                    else:
                        tolman_angle = None

                        # for bidentate ligands with multiple bridges across, only use atoms that
                        # are closer to the key atom we are looking at right now
                        if len(key) == 2:
                            if bridge_path:
                                if key_atom is key1:
                                    other_key = key2
                                else:
                                    other_key = key1
                                frag = self.find(
                                    frag,
                                    CloserTo(
                                        key_atom, other_key, include_ties=True
                                    ),
                                )

                        # some ligands like DuPhos have rings on the phosphorous atom
                        # we only want ones that are closer to the the substituent end
                        frag = self.find(frag, CloserTo(bonded_atom, key_atom))

                        # Geometry(frag).write(outfile="frag%s.xyz" % bonded_atom.name)

                        for atom in frag:
                            beta = np.arcsin(
                                radii_dict[atom.element] / atom.dist(center)
                            )
                            v = center.bond(atom) / center.dist(atom)
                            c = np.linalg.norm(v - L_axis)
                            test_angle = beta + np.arccos((c ** 2 - 2) / -2)

                            if (
                                tolman_angle is None
                                or test_angle > tolman_angle
                            ):
                                tolman_angle = test_angle

                    scale = 5 * np.cos(tolman_angle)

                    cones.append(
                        (
                            center.coords + scale * L_axis,
                            center.coords,
                            scale * abs(np.tan(tolman_angle)),
                        )
                    )

                    total_angle += 2 * tolman_angle / len(bonded_atoms)

            if return_cones:
                return np.rad2deg(total_angle), cones

            return np.rad2deg(total_angle)

        elif method.lower() == "exact":
            beta = np.zeros(len(self.atoms), dtype=float)

            test_one_atom_axis = None
            max_beta = None
            for i, atom in enumerate(self.atoms):
                beta[i] = np.arcsin(
                    radii_dict[atom.element] / atom.dist(center)
                )
                if max_beta is None or beta[i] > max_beta:
                    max_beta = beta[i]
                    test_one_atom_axis = center.bond(atom)

            # check to see if all other atoms are in the shadow of one atom
            # e.g. cyano, carbonyl
            overshadowed_list = []
            for i, atom in enumerate(self.atoms):
                rhs = beta[i]

                if (
                    np.dot(center.bond(atom), test_one_atom_axis)
                    / (center.dist(atom) * np.linalg.norm(test_one_atom_axis))
                    <= 1
                ):
                    rhs += np.arccos(
                        np.dot(center.bond(atom), test_one_atom_axis)
                        / (
                            center.dist(atom)
                            * np.linalg.norm(test_one_atom_axis)
                        )
                    )
                lhs = max_beta
                if lhs >= rhs:
                    # print(atom, "is overshadowed")
                    overshadowed_list.append(atom)
                    break

            # all atoms are in the cone - we're done
            if len(overshadowed_list) == len(self.atoms):
                scale = 5 * np.cos(max_beta)

                cones.append(
                    (
                        center.coords + scale * test_one_atom_axis,
                        center.coords,
                        scale
                        * abs(
                            np.linalg.norm(test_one_atom_axis)
                            * np.tan(max_beta)
                        ),
                    )
                )

                if return_cones:
                    return np.rad2deg(2 * max_beta), cones

                return np.rad2deg(2 * max_beta)

            overshadowed_list = []
            for i, atom1 in enumerate(self.atoms):
                for j, atom2 in enumerate(self.atoms[:i]):
                    rhs = beta[i]

                    if (
                        np.dot(center.bond(atom1), center.bond(atom2))
                        / (center.dist(atom1) * center.dist(atom2))
                        <= 1
                    ):
                        rhs += np.arccos(
                            np.dot(center.bond(atom1), center.bond(atom2))
                            / (center.dist(atom1) * center.dist(atom2))
                        )
                    lhs = beta[j]
                    if lhs >= rhs:
                        overshadowed_list.append(atom1)
                        break
            # winow list to ones that aren't in the shadow of another
            atom_list = [
                atom for atom in self.atoms if atom not in overshadowed_list
            ]

            # check pairs of atoms
            max_a = None
            aij = None
            bij = None
            cij = None
            for i, atom1 in enumerate(atom_list):
                ndx_i = self.atoms.index(atom1)
                for j, atom2 in enumerate(atom_list[:i]):
                    ndx_j = self.atoms.index(atom2)
                    beta_ij = np.arccos(
                        np.dot(center.bond(atom1), center.bond(atom2))
                        / (atom1.dist(center) * atom2.dist(center))
                    )

                    test_alpha = (beta[ndx_i] + beta[ndx_j] + beta_ij) / 2
                    if max_a is None or test_alpha > max_a:
                        max_a = test_alpha
                        mi = center.bond(atom1)
                        mi /= np.linalg.norm(mi)
                        mj = center.bond(atom2)
                        mj /= np.linalg.norm(mj)

                        aij = np.sin(
                            0.5 * (beta_ij + beta[ndx_i] - beta[ndx_j])
                        ) / np.sin(beta_ij)
                        bij = np.sin(
                            0.5 * (beta_ij - beta[ndx_i] + beta[ndx_j])
                        ) / np.sin(beta_ij)
                        cij = 0
                        norm = (
                            aij * mi
                            + bij * mj
                            + cij * np.cross(mi, mj) / np.sin(bij)
                        )

            # r = 0.2 * np.tan(max_a)
            # print(
            #     ".cone %.3f %.3f %.3f   0.0 0.0 0.0   %.3f open" % (
            #         0.2 * norm[0], 0.2 * norm[1], 0.2 * norm[2], r
            #     )
            # )

            overshadowed_list = []
            rhs = max_a
            for atom in atom_list:
                ndx_i = self.atoms.index(atom)
                lhs = beta[ndx_i] + np.arccos(
                    np.dot(center.bond(atom), norm) / center.dist(atom)
                )
                # this should be >=, but there can be numerical issues
                if rhs > lhs or np.isclose(rhs, lhs):
                    overshadowed_list.append(atom)

            # the cone fits all atoms, we're done
            if len(overshadowed_list) == len(atom_list):
                scale = 5 * np.cos(max_a)

                cones.append(
                    (
                        center.coords + (scale * norm),
                        center.coords,
                        scale * abs(np.tan(max_a)),
                    )
                )

                if return_cones:
                    return np.rad2deg(2 * max_a), cones

                return np.rad2deg(2 * max_a)

            centroid = self.COM()
            c_vec = centroid - center.coords
            c_vec /= np.linalg.norm(c_vec)

            min_alpha = None
            c = 0
            for i, atom1 in enumerate(atom_list):
                for j, atom2 in enumerate(atom_list[:i]):
                    for k, atom3 in enumerate(atom_list[i + 1 :]):
                        c += 1
                        ndx_i = self.atoms.index(atom1)
                        ndx_j = self.atoms.index(atom2)
                        ndx_k = self.atoms.index(atom3)
                        # print(atom1.name, atom2.name, atom3.name)

                        mi = center.bond(atom1)
                        mi /= np.linalg.norm(center.dist(atom1))
                        mj = center.bond(atom2)
                        mj /= np.linalg.norm(center.dist(atom2))
                        mk = center.bond(atom3)
                        mk /= np.linalg.norm(center.dist(atom3))

                        gamma_ijk = np.dot(mi, np.cross(mj, mk))

                        # M = np.column_stack((mi, mj, mk))

                        # N = gamma_ijk * np.linalg.inv(M)

                        N = np.column_stack(
                            (
                                np.cross(mj, mk),
                                np.cross(mk, mi),
                                np.cross(mi, mj),
                            )
                        )

                        u = np.array(
                            [
                                np.cos(beta[ndx_i]),
                                np.cos(beta[ndx_j]),
                                np.cos(beta[ndx_k]),
                            ]
                        )

                        v = np.array(
                            [
                                np.sin(beta[ndx_i]),
                                np.sin(beta[ndx_j]),
                                np.sin(beta[ndx_k]),
                            ]
                        )

                        P = np.dot(N.T, N)

                        A = np.dot(u.T, np.dot(P, u))
                        B = np.dot(v.T, np.dot(P, v))
                        C = np.dot(u.T, np.dot(P, v))

                        D = gamma_ijk ** 2

                        # beta_ij = np.dot(center.bond(atom1), center.bond(atom2))
                        # beta_ij /= atom1.dist(center) * atom2.dist(center)
                        # beta_ij = np.arccos(beta_ij)
                        # beta_jk = np.dot(center.bond(atom2), center.bond(atom3))
                        # beta_jk /= atom2.dist(center) * atom3.dist(center)
                        # beta_jk = np.arccos(beta_jk)
                        # beta_ik = np.dot(center.bond(atom1), center.bond(atom3))
                        # beta_ik /= atom1.dist(center) * atom3.dist(center)
                        # beta_ik = np.arccos(beta_ik)
                        #
                        # D = 1 - np.cos(beta_ij) ** 2 - np.cos(beta_jk) ** 2 - np.cos(beta_ik) ** 2
                        # D += 2 * np.cos(beta_ik) * np.cos(beta_jk) * np.cos(beta_ij)
                        # this should be equal to the other D

                        t1 = (A - B) ** 2 + 4 * C ** 2
                        t2 = 2 * (A - B) * (A + B - 2 * D)
                        t3 = (A + B - 2 * D) ** 2 - 4 * C ** 2

                        w_lt = (-t2 - np.sqrt(t2 ** 2 - 4 * t1 * t3)) / (
                            2 * t1
                        )
                        w_gt = (-t2 + np.sqrt(t2 ** 2 - 4 * t1 * t3)) / (
                            2 * t1
                        )

                        alpha1 = np.arccos(w_lt) / 2
                        alpha2 = (2 * np.pi - np.arccos(w_lt)) / 2
                        alpha3 = np.arccos(w_gt) / 2
                        alpha4 = (2 * np.pi - np.arccos(w_gt)) / 2

                        for alpha in [alpha1, alpha2, alpha3, alpha4]:
                            if alpha < max_a:
                                continue

                            if min_alpha is not None and alpha >= min_alpha:
                                continue

                            lhs = (
                                A * np.cos(alpha) ** 2 + B * np.sin(alpha) ** 2
                            )
                            lhs += 2 * C * np.sin(alpha) * np.cos(alpha)
                            if not np.isclose(lhs, D):
                                continue

                            # print(lhs, D)

                            p = np.dot(
                                N, u * np.cos(alpha) + v * np.sin(alpha)
                            )
                            norm = p / gamma_ijk

                            for atom in atom_list:
                                ndx = self.atoms.index(atom)
                                rhs = beta[ndx]
                                d = np.dot(
                                    center.bond(atom), norm
                                ) / center.dist(atom)
                                if abs(d) < 1:
                                    rhs += np.arccos(d)
                                if not alpha >= rhs:
                                    break
                            else:
                                if min_alpha is None or alpha < min_alpha:
                                    # print("min_alpha set", alpha)
                                    min_alpha = alpha
                                    min_norm = norm
                                    # r = 2 * np.tan(min_alpha)
                                    # print(
                                    #     ".cone %.3f %.3f %.3f   0.0 0.0 0.0   %.3f open" % (
                                    #         2 * norm[0], 2 * norm[1], 2 * norm[2], r
                                    #     )
                                    # )

            scale = 5 * np.cos(min_alpha)

            cones.append(
                (
                    center.coords + scale * min_norm,
                    center.coords,
                    scale * abs(np.tan(min_alpha)),
                )
            )

            if return_cones:
                return np.rad2deg(2 * min_alpha), cones

            return np.rad2deg(2 * min_alpha)

        else:
            raise NotImplementedError(
                "cone angle type is not implemented: %s" % method
            )



#!/usr/bin/env python3
# testing for command line scripts
import os
import ssl
import sys
import unittest
from glob import glob
from subprocess import PIPE, Popen

import AaronTools
from AaronTools.atoms import Atom
from AaronTools.fileIO import FileReader
from AaronTools.geometry import CACTUS_HOST, OPSIN_HOST, Geometry
from AaronTools.test import TestWithTimer, prefix, rmsd_tol, validate
from AaronTools.test.test_geometry import is_close


class TestCLS(TestWithTimer):
    benz_NO2_Cl = os.path.join(prefix, "test_files", "benzene_1-NO2_4-Cl.xyz")

    benzene = os.path.join(prefix, "test_files", "benzene.xyz")

    pyridine = os.path.join(prefix, "test_files", "pyridine.xyz")

    chlorotoluene = os.path.join(prefix, "test_files", "chlorotoluene.xyz")
    chlorotoluene_ref = os.path.join(
        prefix, "ref_files", "chlorotoluene_180.xyz"
    )

    benzene_dimer = os.path.join(prefix, "test_files", "benzene_dimer.xyz")
    benzene_dimer_ref = os.path.join(
        prefix, "ref_files", "benzene_dimer_ref.xyz"
    )

    naphthalene = os.path.join(prefix, "ref_files", "naphthalene.xyz")

    tetrahydronaphthalene = os.path.join(
        prefix, "ref_files", "tetrahydronaphthalene.xyz"
    )

    pyrene = os.path.join(prefix, "ref_files", "pyrene.xyz")

    benz_OH_Cl = os.path.join(prefix, "test_files", "benzene_1-OH_4-Cl.xyz")

    frequencies = os.path.join(prefix, "test_files", "normal.log")

    rmsd_sort_1 = os.path.join(prefix, "test_files", "test_rmsd_sort1.xyz")
    rmsd_sort_2 = os.path.join(prefix, "test_files", "test_rmsd_sort2.xyz")

    g09_com_file = os.path.join(
        prefix, "test_files", "5a-sub1.R.ts1.Cf1.3.com"
    )
    g09_log_file = os.path.join(prefix, "test_files", "opt_normal.log")
    orca_out_file = os.path.join(prefix, "test_files", "orca_geom.out")
    psi4_dat_file = os.path.join(prefix, "test_files", "psi4-test.out")
    xyz_file = os.path.join(prefix, "test_files", "benzene.xyz")

    tm_simple = os.path.join(
        prefix, "test_files", "catalysts", "tm_single-lig.xyz"
    )

    t60 = os.path.join(prefix, "test_files", "torsion-60.xyz")
    t90 = os.path.join(prefix, "test_files", "torsion-90.xyz")

    opt_file_1 = os.path.join(prefix, "test_files", "opt_running.log")

    make_conf_1 = os.path.join(prefix, "test_files", "R-Quinox-tBu3.xyz")
    make_conf_ref_1 = sorted(
        glob(os.path.join(prefix, "ref_files", "make_conf_cls", "*.xyz"))
    )

    change_chir_1 = os.path.join(prefix, "test_files", "chiral_ring.xyz")
    change_chir_ref_1 = sorted(
        glob(
            os.path.join(prefix, "ref_files", "change_chirality_cls", "*.xyz")
        )
    )
    chiral_ring_mirror = os.path.join(
        prefix, "test_files", "chiral_ring_mirror.xyz"
    )
    cone_bidentate_2 = os.path.join(prefix, "test_files", "bpy.xyz")
    cone_bidentate_3 = os.path.join(prefix, "test_files", "dppe.xyz")

    aarontools_bin = os.path.join(os.path.dirname(AaronTools.__file__), "bin")
    # CLS stored in $PYTHONHOME/bin if installed via pip
    if aarontools_bin not in os.getenv("PATH"):
        proc = Popen(["pip", "show", "AaronTools"], stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()
        if out:
            tmp = os.path.split(aarontools_bin)
            while tmp[1] and tmp[1] != "lib":
                tmp = os.path.split(tmp[0])
            aarontools_bin = os.path.join(tmp[0], "bin")

    def test_environment(self):
        """is this AaronTools' bin in the path?"""
        path = os.getenv("PATH")
        self.assertTrue(self.aarontools_bin in path)

    # geometry measurement
    def test_angle(self):
        """measuring angles"""
        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "angle.py"),
            TestCLS.benz_NO2_Cl,
            "-m",
            "13",
            "12",
            "14",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        angle = float(out)
        self.assertTrue(is_close(angle, 124.752, 10 ** -2))

    def test_bond(self):
        """measuring bonds"""
        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "bond.py"),
            TestCLS.benzene,
            "-m",
            "1",
            "2",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        angle = float(out)
        self.assertTrue(is_close(angle, 1.3952, 10 ** -2))

    def test_dihedral(self):
        """measuring dihedrals"""
        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "dihedral.py"),
            TestCLS.benz_NO2_Cl,
            "-m",
            "13",
            "12",
            "1",
            "6",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        dihedral = float(out)
        self.assertTrue(is_close(dihedral, 45.023740, 10 ** -5))

    def test_rmsdAlign(self):
        """measuring rmsd"""
        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "rmsdAlign.py"),
            "-r",
            TestCLS.benz_NO2_Cl,
            TestCLS.benz_NO2_Cl,
            "--value",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        rmsd = float(out)
        self.assertTrue(is_close(rmsd, 0, 10 ** -5))

        # test sorting flag
        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "rmsdAlign.py"),
            "-r",
            TestCLS.rmsd_sort_1,
            TestCLS.rmsd_sort_2,
            "--value",
            "--sort",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        rmsd = float(out)
        self.assertTrue(rmsd < 0.1)

    def test_substitute(self):
        """test substitute.py"""
        ref = Geometry(TestCLS.benz_NO2_Cl)

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "substitute.py"),
            TestCLS.benzene,
            "-s",
            "12=NO2",
            "-s",
            "11=Cl",
        ]
        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()
        # print(out.decode("utf-8"))
        # print(err.decode("utf-8"))
        fr = FileReader(("out", "xyz", out.decode("utf-8")))
        mol = Geometry(fr)
        self.assertTrue(validate(mol, ref))

        # I don't want to run these tests anymore
        # sometimes the server timeout stuff doesn't seem to work and
        # the test hangs for > 30 seconds
        # - Tony
        # # don't run smiles/iupac tests if we can't connect to the host site
        # try:
        #     for host in [CACTUS_HOST, OPSIN_HOST]:
        #         response = urllib.request.urlopen(
        #             host, context=ssl.SSLContext()
        #         )
        #         # name resolved, but something wrong with server
        #         if response.status >= 5:
        #             return
        # except urllib.error.URLError:
        #     # can't resolve name or some other connectivity error
        #     return
        #
        # args = [
        #     sys.executable,
        #     os.path.join(self.aarontools_bin, "substitute.py"),
        #     TestCLS.benzene,
        #     "-s",
        #     "12=smiles:O=[N.]=O",
        #     "-s",
        #     "11=iupac:chloro",
        # ]
        # proc = Popen(args, stdout=PIPE, stderr=PIPE)
        # out, err = proc.communicate()
        # fr = FileReader(("out", "xyz", out.decode("utf-8")))
        # mol = Geometry(fr)
        # self.assertTrue(validate(mol, ref, thresh=2e-1))

    def test_closeRing(self):
        """test closeRing.py"""
        ref1 = Geometry(TestCLS.naphthalene)

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "closeRing.py"),
            TestCLS.benzene,
            "-r",
            "7",
            "8",
            "benzene",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        fr = FileReader(("out", "xyz", out.decode("utf-8")))
        mol = Geometry(fr)
        rmsd = mol.RMSD(ref1, sort=True)
        self.assertTrue(rmsd < rmsd_tol(ref1, superLoose=True))

    def test_mirror(self):
        """test mirror.py"""
        ref1 = Geometry(TestCLS.chiral_ring_mirror)

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "mirror.py"),
            TestCLS.change_chir_1,
            "-xy",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        fr = FileReader(("out", "xyz", out.decode("utf-8")))
        mol = Geometry(fr)
        rmsd = mol.RMSD(ref1, sort=True)
        self.assertTrue(rmsd < rmsd_tol(ref1, superLoose=True))

    def test_grabThermo(self):
        """test grabThermo.py"""
        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "grabThermo.py"),
            TestCLS.frequencies,
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        ref = """electronic energy of test_files/normal.log = -1856.018658 Eh
    E+ZPE             = -1855.474686 Eh  (ZPE = 0.543972)
thermochemistry from test_files/normal.log at 298.00 K:
    H(RRHO)           = -1855.440616 Eh  (dH = 0.578042)
    G(RRHO)           = -1855.538017 Eh  (dG = 0.480642)
  quasi treatments for entropy (w0=100.0 cm^-1):
    G(Quasi-RRHO)     = -1855.532805 Eh  (dG = 0.485854)
    G(Quasi-harmonic) = -1855.532510 Eh  (dG = 0.486148)
"""
        # strip b/c windows adds \r to the end of lines
        out_list = out.decode("utf-8").splitlines()
        ref_list = ref.splitlines()

        # can't test all the lines b/c paths might be different
        # test sp energy
        self.assertTrue(out_list[0][-16:] == ref_list[0][-16:])
        # test thermochem
        for i in [1, 3, 4, 6, 7]:
            self.assertTrue(out_list[i][-34:] == ref_list[i][-34:])

        # test regular output with sp
        # sp is the same as the thermo file
        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "grabThermo.py"),
            TestCLS.frequencies,
            "-sp",
            TestCLS.frequencies,
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        out_list = out.decode("utf-8").splitlines()
        self.assertTrue(out_list[0][-16:] == ref_list[0][-16:])
        for i in [1, 3, 4, 6, 7]:
            self.assertTrue(out_list[i][-34:] == ref_list[i][-34:])

        # test CSV w/o sp file
        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "grabThermo.py"),
            TestCLS.frequencies,
            "-csv",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        ref_csv = """E,E+ZPE,H(RRHO),G(RRHO),G(Quasi-RRHO),G(Quasi-harmonic),ZPE,dH(RRHO),dG(RRHO),dG(Quasi-RRHO),dG(Quasi-harmonic),SP_File,Thermo_File
-1856.018658,-1855.474686,-1855.440616,-1855.538017,-1855.532805,-1855.532510,0.543972,0.578042,0.480642,0.485854,0.486148,test_files/normal.log,test_files/normal.log"""

        out_list = out.decode("utf-8").splitlines()
        ref_list = ref_csv.splitlines()

        out_list = out.decode("utf-8").splitlines()
        ref_list = ref_csv.splitlines()

        # test CSV with sp file
        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "grabThermo.py"),
            TestCLS.frequencies,
            "-csv",
            "-sp",
            TestCLS.frequencies,
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        out_list = out.decode("utf-8").splitlines()
        ref_list = ref_csv.splitlines()

        self.assertTrue(out_list[0] == ref_list[0])
        self.assertTrue(
            out_list[1].split(",")[:-2] == ref_list[1].split(",")[:-2]
        )

        # test CSV with looking in subdirectories
        filename = os.path.basename(TestCLS.frequencies)
        directory = os.path.join(prefix, "test_files")
        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "grabThermo.py"),
            directory,
            "-r",
            filename,
            "-csv",
            "-sp",
            filename,
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        out_list = out.decode("utf-8").splitlines()
        ref_list = ref_csv.splitlines()

        self.assertTrue(out_list[0] == ref_list[0])
        self.assertTrue(
            out_list[1].split(",")[:-2] == ref_list[1].split(",")[:-2]
        )

    def test_printXYZ(self):
        """test printXYZ.py"""
        # for each test, the rmsd tolerance is determined based on the number of atoms and
        # the precision we use when printing xyz files
        # test xyz file
        ref_xyz = Geometry(TestCLS.xyz_file)

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "printXYZ.py"),
            TestCLS.xyz_file,
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        fr = FileReader(("out", "xyz", out.decode("utf-8")))
        mol = Geometry(fr)
        rmsd = mol.RMSD(ref_xyz, align=True)
        self.assertTrue(rmsd < len(ref_xyz.atoms) * (3 * 1e-5))

        # test gaussian input file
        ref_com = Geometry(TestCLS.g09_com_file)

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "printXYZ.py"),
            TestCLS.g09_com_file,
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        fr = FileReader(("out", "xyz", out.decode("utf-8")))
        mol = Geometry(fr)
        rmsd = mol.RMSD(ref_com, align=True)
        self.assertTrue(rmsd < len(ref_com.atoms) * (3 * 1e-5))

        # test gaussian output file
        ref_log = Geometry(TestCLS.g09_log_file)

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "printXYZ.py"),
            TestCLS.g09_log_file,
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        fr = FileReader(("out", "xyz", out.decode("utf-8")))
        mol = Geometry(fr)
        rmsd = mol.RMSD(ref_log, align=True)
        self.assertTrue(rmsd < len(ref_log.atoms) * (3 * 1e-5))

        # test orca output file
        ref_out = Geometry(TestCLS.orca_out_file)

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "printXYZ.py"),
            TestCLS.orca_out_file,
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        fr = FileReader(("out", "xyz", out.decode("utf-8")))
        mol = Geometry(fr)
        rmsd = mol.RMSD(ref_out, align=True)
        self.assertTrue(rmsd < len(ref_out.atoms) * (3 * 1e-5))

        # test psi4 output files and format flat
        ref_dat = Geometry(FileReader((TestCLS.psi4_dat_file, "dat", None)))

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "printXYZ.py"),
            TestCLS.psi4_dat_file,
            "-if",
            "dat",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        fr = FileReader(("out", "xyz", out.decode("utf-8")))
        mol = Geometry(fr)
        rmsd = mol.RMSD(ref_dat, align=True)
        self.assertTrue(rmsd < len(ref_dat.atoms) * (3 * 1e-5))

    def test_changeElement(self):
        """test changeElement.py"""
        ref = Geometry(TestCLS.pyridine)

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "changeElement.py"),
            TestCLS.benzene,
            "-e",
            "1=N",
            "-c",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            print(out.decode("utf-8"))
            print(err.decode("utf-8"))
            raise RuntimeError(err.decode("utf-8"))

        fr = FileReader(("out", "xyz", out.decode("utf-8")))
        mol = Geometry(fr)
        self.assertTrue(validate(ref, mol, thresh="loose"))

    def test_rotate(self):
        """test rotate.py"""
        ref = Geometry(TestCLS.chlorotoluene_ref)

        # range of targets
        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "rotate.py"),
            TestCLS.chlorotoluene,
            "-b",
            "3",
            "12",
            "-c",
            "3",
            "-t",
            "12-15",
            "-a",
            "180",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        fr = FileReader(("out", "xyz", out.decode("utf-8")))
        mol = Geometry(fr)
        rmsd = mol.RMSD(ref, align=True)
        self.assertTrue(rmsd < rmsd_tol(ref))

        # enumerate all targets
        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "rotate.py"),
            TestCLS.chlorotoluene,
            "-b",
            "3",
            "12",
            "-c",
            "3",
            "-t",
            "12,13,14,15",
            "-a",
            "180",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        fr = FileReader(("out", "xyz", out.decode("utf-8")))
        mol = Geometry(fr)
        rmsd = mol.RMSD(ref, align=True)
        self.assertTrue(rmsd < rmsd_tol(ref))

        # rotate all atom by 180 - rmsd should be basically 0
        ref2 = Geometry(TestCLS.chlorotoluene)

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "rotate.py"),
            TestCLS.chlorotoluene,
            "-x",
            "x",
            "-a",
            "180",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        fr = FileReader(("out", "xyz", out.decode("utf-8")))
        mol = Geometry(fr)
        rmsd = mol.RMSD(ref2, align=True)
        self.assertTrue(rmsd < rmsd_tol(ref2))

        # rotate one fragment
        ref3 = Geometry(TestCLS.benzene_dimer_ref)

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "rotate.py"),
            TestCLS.benzene_dimer,
            "-p",
            "1-12",
            "-c",
            "1-12",
            "-f",
            "1",
            "-a",
            "10",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        fr = FileReader(("out", "xyz", out.decode("utf-8")))
        mol = Geometry(fr)
        rmsd = mol.RMSD(ref3, align=True)
        self.assertTrue(rmsd < rmsd_tol(ref3))

    def test_mapLigand(self):
        """test mapLigand.py"""
        ref = Geometry(os.path.join(prefix, "ref_files", "lig_map_3.xyz"))

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "mapLigand.py"),
            TestCLS.tm_simple,
            "-l",
            "35,36=S-tBu-BOX",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        fr = FileReader(("out", "xyz", out.decode("utf-8")))
        mol = Geometry(fr)
        self.assertTrue(validate(mol, ref, thresh="loose", debug=False))

    def test_interpolate(self):
        """test interpolate.py
        assumes current working directory is writable b/c interpolate doesn't
        print structures to stdout"""
        ref = Geometry(
            os.path.join(prefix, "ref_files", "torsion_interpolation.xyz")
        )

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "interpolate.py"),
            TestCLS.t60,
            TestCLS.t90,
            "-t",
            "0.40",
            "-u",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        mol = Geometry("traj-0.xyz")
        os.remove("traj-0.xyz")
        rmsd = mol.RMSD(ref, align=True, sort=True)
        self.assertTrue(rmsd < rmsd_tol(ref, superLoose=True))

    def test_makeConf(self):
        """test makeConf.py"""

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "makeConf.py"),
            TestCLS.make_conf_1,
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        fr = FileReader(("out", "xyz", out.decode("utf-8")), get_all=True)

        for step, ref in zip(fr.all_geom, self.make_conf_ref_1):
            geom = None
            for item in step:
                if isinstance(item, list) and all(
                    isinstance(a, Atom) for a in item
                ):
                    geom = Geometry(item)

            if geom is None:
                raise RuntimeError("an output is missing atoms")

            ref_geom = Geometry(ref)
            rmsd = ref_geom.RMSD(geom)
            self.assertTrue(rmsd < rmsd_tol(ref_geom))

    def test_changeChirality(self):
        """test changeChirality.py"""

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "changeChirality.py"),
            TestCLS.change_chir_1,
            "--diastereomers",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        fr = FileReader(("out", "xyz", out.decode("utf-8")), get_all=True)

        for step, ref in zip(fr.all_geom, self.change_chir_ref_1):
            geom = None
            for item in step:
                if isinstance(item, list) and all(
                    isinstance(a, Atom) for a in item
                ):
                    geom = Geometry(item)

            if geom is None:
                raise RuntimeError("an output is missing atoms")

            ref_geom = Geometry(ref)
            rmsd = ref_geom.RMSD(geom)
            self.assertTrue(rmsd < rmsd_tol(ref_geom))

    def test_grabStatus(self):
        """test grabStatus.py"""

        # gaussian file that is not finished
        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "grabStatus.py"),
            TestCLS.opt_file_1,
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        # ref = """                      Filename        Max Disp       Max Force        RMS Disp       RMS Force
        #     test_files/opt_running.log     1.66e+00/NO     3.61e+00/NO     2.79e-01/NO     2.55e-01/NO  not finished"""

        ref = """                      Filename    Step        Max Disp       Max Force        RMS Disp       RMS Force
            test_files/opt_running.log       4     1.66e+00/NO     3.61e+00/NO     2.79e-01/NO     2.55e-01/NO  not finished"""

        ref_lines = ref.splitlines()
        ref_status_line = ref_lines[1]

        lines = out.decode("utf-8").splitlines()
        test_line = lines[1]

        # don't include filename in test b/c that will be different
        for ref_item, test_item in zip(
            ref_status_line.split()[-7:], test_line.split()[-7:]
        ):
            # print(ref_item, test_item)
            self.assertTrue(ref_item == test_item)

    def test_substituentSterimol(self):
        """test substituentSterimol.py"""

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "substituentSterimol.py"),
            TestCLS.benzene,
            "-s", "1",
            "-a" "12",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        ref = """B1      B2      B3      B4      B5      L       file
        1.70    1.70    3.36    3.36    3.36    6.78    test_files/benzene.xyz

        """

        ref_lines = ref.splitlines()
        ref_status_line = ref_lines[1]

        lines = out.decode("utf-8").splitlines()
        test_line = lines[1]

        # don't include filename in test b/c that will be different
        for ref_item, test_item in zip(
            ref_status_line.split()[:6], test_line.split()[:6]
        ):
            if ref_item != test_item:
                print(ref_item, test_item)
            self.assertTrue(ref_item == test_item)

        # sterimol2vec test
        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "substituentSterimol.py"),
            TestCLS.benzene,
            "-s", "1",
            "-a" "12",
            "-al", "2,2.1,2.2,2.3,2.4,2.5,2.6,2.7,2.8,2.9,3"
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        if len(err) != 0:
            raise RuntimeError(err)

        ref = """B1      B2      B3      B4      B5      L       file
        1.67    1.67    3.35    3.35    3.35    2.00    test_files\benzene.xyz
        1.67    1.67    3.33    3.33    3.33    2.10    test_files\benzene.xyz
        1.69    1.69    3.31    3.31    3.31    2.20    test_files\benzene.xyz
        1.70    1.70    3.27    3.27    3.27    2.30    test_files\benzene.xyz
        1.70    1.70    3.23    3.23    3.23    2.40    test_files\benzene.xyz
        1.70    1.70    3.17    3.17    3.17    2.50    test_files\benzene.xyz
        1.69    1.69    3.09    3.10    3.10    2.60    test_files\benzene.xyz
        1.67    1.67    3.00    3.01    3.01    2.70    test_files\benzene.xyz
        1.65    1.65    2.89    2.89    2.89    2.80    test_files\benzene.xyz
        1.62    1.62    2.83    2.83    2.83    2.90    test_files\benzene.xyz
        1.59    1.59    2.80    2.80    2.80    3.00    test_files\benzene.xyz
        
        """
        
        for ref_line, test_line in zip(out.decode("utf-8").splitlines(), ref.splitlines()):
            for ref_val, test_val in zip(ref_line.split()[:-1], test_line.split()[:-1]):
                if ref_val != test_val:
                    print(ref_line, test_line)
                self.assertEqual(ref_val, test_val)

    def test_coneAngle(self):
        """test coneAngle.py"""

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "coneAngle.py"),
            "-r",
            "bondi",
            TestCLS.cone_bidentate_3,
            "-k",
            "2,3",
            "-m",
            "exact",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        angle = float(out)
        self.assertTrue(abs(angle - 218.6) <= 0.1)

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "coneAngle.py"),
            "-r",
            "bondi",
            TestCLS.cone_bidentate_2,
            "-k",
            "2,3",
            "-m",
            "exact",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        angle = float(out)
        self.assertTrue(abs(angle - 194.6) <= 0.1)

    def test_ligandSterimol(self):
        """test ligandSterimol.py"""

        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "ligandSterimol.py"),
            "-r", "umn",
            TestCLS.tm_simple,
            "-k", "35,36",
            "-c", "34",
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        ref = """B1      B2      B3      B4      B5      L       file
        3.98    5.34    5.63    6.28    6.31    9.65    test_files\catalysts\tm_single-lig.xyz
        
        """

        for ref_line, test_line in zip(out.decode("utf-8").splitlines(), ref.splitlines()):
            for ref_val, test_val in zip(ref_line.split()[:-1], test_line.split()[:-1]):
                if ref_val != test_val:
                    print(ref_line, test_line)
                self.assertEqual(ref_val, test_val)

        # Sterimol2Vec
        args = [
            sys.executable,
            os.path.join(self.aarontools_bin, "ligandSterimol.py"),
            "-r", "umn",
            TestCLS.tm_simple,
            "-k", "35,36",
            "-c", "34",
            "-al", "2.5,3,3.5,4"
        ]

        proc = Popen(args, stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()
        
        ref = """B1      B2      B3      B4      B5      L       file
        3.94    4.81    5.36    6.27    6.28    2.50    test_files\catalysts\tm_single-lig.xyz
        3.75    4.63    5.12    6.05    6.05    3.00    test_files\catalysts\tm_single-lig.xyz
        3.18    4.08    5.33    5.44    5.60    3.50    test_files\catalysts\tm_single-lig.xyz
        2.39    3.85    4.71    4.95    5.41    4.00    test_files\catalysts\tm_single-lig.xyz
        
        """

        for ref_line, test_line in zip(out.decode("utf-8").splitlines(), ref.splitlines()):
            for ref_val, test_val in zip(ref_line.split()[:-1], test_line.split()[:-1]):
                if ref_val != test_val:
                    print(ref_line, test_line)
                self.assertEqual(ref_val, test_val)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestCLS("test_substituentSterimol"))
    suite.addTest(TestCLS("test_ligandSterimol"))
    return suite


ONLYSOME = False

if __name__ == "__main__" and ONLYSOME:
    runner = unittest.TextTestRunner()
    runner.run(suite())
elif __name__ == "__main__":
    unittest.main()
